from pycarol.pipeline.utils.hash_versioning.test_hash_versioning import (
    equal_functions_list, different_functions_list
)

function_set = set()
for a,b in equal_functions_list + different_functions_list:
    function_set.add(a)
    function_set.add(b)

function_list = [f for f in function_set]

from pycarol.pipeline.utils.hash_versioning import get_bytecode_tree, get_function_hash
def get_hash(f):
    name = f.__name__

    try:
        bytecode = get_bytecode_tree(f)
    except:
        bytecode = "FAIL"
    try:
        h = get_function_hash(f)
    except:
        h = "FAIL"
    
    return([name, bytecode, h])

def get_hash_all(f_list,q):
    for f in f_list:
        q.put(get_hash(f))

def print_all():
    for f in function_set:
        name, bytecode, h = get_hash(f)
        print(name)
        print(bytecode)
        print(h)


import multiprocessing as mp

from multiprocessing import Process, Queue
import time
import os


def queue_to_list(q)->list:
    l=[]
    while not q.empty():
        l.append(q.get())
    return l

if __name__ == '__main__':
    mp.get_context('spawn')
    q1 = Queue()
    q2 = Queue()
    p1 = Process(target=get_hash_all, args=(function_list,q1,))
    p2 = Process(target=get_hash_all, args=(function_list,q2,))
    p1.start()
    p1.join()
    p2.start()
    p2.join()
    
    l1 = queue_to_list(q1)
    l2 = queue_to_list(q2)
    assert l1==l2
    names1 = [n[0] for n in l1]
    names2 = [n[0] for n in l2]
    hash1 = [n[2] for n in l1]
    hash2 = [n[2] for n in l2]

    for i, name in enumerate(names1):
        print(name)
        print('\t',hash1[i])
        print('\t',hash2[i])
        print('\t',hash1[i]==hash2[i])



class Organization:
    """
    Get organization information.

    Args:

        Carol object
            Carol object.
    """

    def __init__(self, carol):
        self.carol = carol

    def get_organization_info(self, organization, auth=True):
        """

        Args:
            organization: `str`
                Organization subdomain
            auth: `bool`
            If send the auth information. Some fields do not come if not logged.

        Returns:
             dict with the information about the organization.

        """

        return self.carol.call_api(f'v1/organizations/domain/{organization}',
                                   auth=auth, status_forcelist=[], retries=0)

    def get_environment_info(self, environment, auth=True):
        """
        Get environment information.

        Args:
            environment: `str`
                Tenant name
            auth: `bool`
                If send the auth information. Some fields do not come if not logged.

        Returns:
            dict with the information about the environment.

        """

        return self.carol.call_api(f'v2/tenants/domain/{environment}',
                                   auth=auth, status_forcelist=[], retries=0)

    def get_org_by_id(self, org_id, auth=True):
        """
        Get organization info by id.

        Args:
            org_id: `str`
                Organization Id
            auth: `bool`
                If send the auth information. Some fields do not come if not logged.

        Returns:
            dict with the information about the organization.

        """

        return self.carol.call_api(f'v1/organizations/{org_id}', method='GET', auth=auth)

    def get_env_by_id(self, env_id, auth=True):
        """
        Get env. info by id.

        Args:
            env_id: `str`
                Organization Id
            auth: `bool`
                If send the auth information. Some fields do not come if not logged.

        Returns:
            dict with the information about the organization.

        """

        return self.carol.call_api(f'v2/admin/tenants/{env_id}', method='GET', auth=auth)



"""
Logger Handler integrated with Carol.

"""

from . import Tasks, Carol
import os
import logging
import sys
import re

_carol_levels = dict(
    NOTSET="NOTSET",
    DEBUG='DEBUG',
    INFO="INFO",
    WARNING="WARN",
    WARN="WARN",
    ERROR="ERROR",
    CRITICAL="ERROR",
    FATAL="ERROR",
)


class CarolHandler(logging.StreamHandler):
    """
    Carol logger handler.

    This class can be used to log information in long tasks in Carol.

    Args:

        carol: Carol object
            Carol object.

    Usage:

    .. code:: python

        from pycarol import Carol, CarolHandler
        import logging
        logger = logging.getLogger(__name__)
        carol = CarolHandler(Carol())
        carol.setLevel(logging.INFO)
        logger.addHandler(carol)

        logger.debug('This is a debug message') #This will not be logged in Carol. Level is set to INFO
        logger.info('This is an info message')
        logger.warning('This is a warning message')
        logger.error('This is an error message')
        logger.critical('This is a critical message')

        #These methods will use the current long task id provided by Carol when running your application.
        #For local environments you need to set that manually first on the beginning of your code:

        import os
        os.environ['LONGTASKID'] = TASK_ID

    If no TASK ID is passed it works as a Console Handler.
    """

    def __init__(self, carol=None):
        """

        """
        super().__init__(stream=sys.stdout)

        self._use_console = False
        if carol is None:
            domain = os.getenv('CAROLTENANT')
            app_name = os.getenv('CAROLAPPNAME')
            auth_token = os.getenv('CAROLAPPOAUTH')
            connector_id = os.getenv('CAROLCONNECTORID')
            if ((domain is not None)
                    and (app_name is not None)
                    and (auth_token is not None)
                    and (connector_id is not None)):
                carol = Carol()
                self._use_console = False

            else:
                self._use_console = True

        self.carol = carol
        self._task = Tasks(self.carol)
        self.task_id = os.getenv('LONGTASKID', None)
        self._task.task_id = self.task_id
        self._first_pending = True

    def _log_carol(self, record):
        msg = self.format(record)
        log_level = _carol_levels.get(record.levelname)
        self._task.add_log(msg, log_level=log_level)

    def emit(self, record):
        """
         Log the message.

        Args:
            record: `str`
                Message to log.

        Returns: None

        """

        if (self.task_id is None) or (self._use_console):
            super().emit(record)
        else:
            self._log_carol(record)





def load_pipeline():
    try:
        import app.pipeline
    except ModuleNotFoundError as e:
        print(
            "Module app.pipeline was not found. Please, pip install your app "
            "and define your pipeline in app/pipeline.py")
        raise e

    from importlib import reload
    app_pipeline = reload(app.pipeline)

    from pycarol.pipeline.tools import Pipe

    try:
        pipeline = app_pipeline.pipeline
        assert isinstance(pipeline, Pipe)
    except AttributeError as e:
        print("Please define the object pipeline in app/pipeline.py")
        raise e
    except AssertionError as e:
        print(
            "pipeline object in app/pipeline.py should be an instance of from "
            "pycarol.pipeline.tools.Pipe class")
        raise e

    return pipeline


def show_pipeline():
    pipeline = load_pipeline()
    from pycarol.pipeline.viewer.dash_plot import get_app_from_pipeline
    app = get_app_from_pipeline(pipeline)

    import jupyterlab_dash
    import dash
    import dash_html_components as html
    viewer = jupyterlab_dash.AppViewer()
    viewer.show(app)


from .carolina import Carolina
from datetime import datetime, timedelta

class Storage:

    """
    Handle all Carol storage interactions.


    Args:
            carol: `class: pycarol.Carol`

    """

    def __init__(self, carol):
        """

        Args:

            carol: `class: pycraol.Carol`
        """

        self.carol = carol
        self._init_if_needed()
        self.carolina = None

    def _init_if_needed(self):
        # TODO: This if seems to be useless. It can be handled in `carolina.init_if_needed()`
        if (Carolina.token is not None) and (Carolina.token.get('tenant_name', '') == self.carol.tenant['mdmName']) and \
           (Carolina.token.get('app_name', '') == self.carol.app_name) and \
           (datetime.utcnow() + timedelta(minutes=1) < datetime.fromtimestamp(Carolina.token.get('expirationTimestamp', 1)/1000.0)):
            pass
            #return
        else:
            Carolina.token = None

        self.carolina = Carolina(self.carol)
        self.carolina.init_if_needed()
        if self.carolina.engine == 'GCP-CS':
            from .utils.storage_gcpcs import StorageGCPCS
            self.backend = StorageGCPCS(self.carol, self.carolina)
        else:
            raise NotImplemented(f"Only 'GCP-CS' backend implemented in this version. "
                                 f"You are trying to use {self.carolina.engine }")

    def save(self, name, obj, format='pickle', parquet=False, cache=True, chunk_size=None):
        """

        Args:

            name: `str`.
                Filename to be used when saving the `obj`
            obj: `obj`
                It depends on the `format` parameter.
            format: `str`
                Possible values:

                    1. `pickle`: It uses `pickle.dump` to save the binary file.
                    2. `joblib`: It uses `joblib.dump` to save the binary file.
                    3. `file`: It saves a local file sending it directly to Carol.

            parquet: `bool` default `False`
                It uses `pandas.DataFrame.to_parquet` to save. `obj` should be a pandas DataFrame
            cache: `bool` default `True`
                Cache the file saved in the temp directory.
            chunk_size: `int` default `None`
                The size of a chunk of data whenever iterating (in bytes).
                This must be a multiple of 256 KB per the API specification.

        Usage:

        Saving a local file in CDS.

        .. code:: python

            from pycarol import Carol, Storage
            import pandas as pd
            login = Carol()
            stg = Storage(login)

            stg.save(name='myfile.csv', obj='/local/file/.csv',  format='file')
            # to load the file use:
            path = stg.load(name='teste.zip',  format='file')
            pd.read_csv(path)

        Saving an object.

        .. code:: python

            my_dict = {"a":1, "b":2}

            from pycarol import Carol, Storage
            login = Carol()
            stg = Storage(login)

            stg.save(name='myfile.json', obj=my_dict,  format='pickle')
            # to load the file use:
            my_dict = stg.load(name='myfile.json',  format='pickle')

        It works for `format=joblib` as well,

        .. code:: python

            my_dict = {"a":1, "b":2}

            from pycarol import Carol, Storage
            login = Carol()
            stg = Storage(login)

            stg.save(name='myfile.json', obj=my_dict,  format='joblib')
            # to load the file use:
            my_dict = stg.load(name='myfile.json',  format='joblib')

        Saving a pandas DataFrame

        .. code:: python

            import pandas as pd
            from pycarol import Carol, Storage

            d = {'col1': [1, 2], 'col2': [3, 4]}
            df = pd.DataFrame(data=d)

            login = Carol()
            stg = Storage(login)

            stg.save(name='myfile.parquet', obj=my_dict,  parquet=True)
            # to load the file use:
            df = stg.load(name='myfile.parquet', parquet=True)


        """
        self.backend.save(name, obj, format, parquet, cache, chunk_size)

    def load(self, name, format='pickle', parquet=False, cache=True, storage_space='app_storage', columns=None,
             chunk_size=None):
        """

        Args:

            name: `str`.
                Filename to be load
            format: `str`
                Possible values:

                    1. `pickle`: It uses `pickle.dump` to save the binary file.
                    2. `joblib`: It uses `joblib.dump` to save the binary file.
                    3. `file`: It saves a local file sending it directly to Carol.

            parquet: `bool` default `False`
                It uses `pandas.DataFrame.to_parquet` to save. `obj` should be a pandas DataFrame
            cache: `bool` default `True`
                Cache the file saved in the temp directory.
            storage_space: `str` default `app_storage`
                Internal Storage space.
                    1. "app_storage": For raw storage.
                    2. "golden": Data Model golden records.
                    3. "staging": Staging records path
                    4. "staging_master": Staging records from Master
                    5. "staging_rejected": Staging records from Rejected
                    6. "view": Data Model View records
                    7. "app": App  bucket
                    8. "golden_cds": CDS golden records
                    9. "staging_cds": Staging Intake.
            columns: `list` default `None`
                Columns to fetch when using `parquet=True`
            chunk_size: `int` default `None`
                The size of a chunk of data whenever iterating (in bytes).
                This must be a multiple of 256 KB per the API specification.
        Usage:

        Loading a local file in CDS.

        .. code:: python

            from pycarol import Carol, Storage
            import pandas as pd
            login = Carol()
            stg = Storage(login)

            path = stg.load(name='myfile.csv', format='file')
            pd.read_csv(path)

        Loading an object.

        .. code:: python

            from pycarol import Carol, Storage
            login = Carol()
            stg = Storage(login)

            my_dict = stg.load(name='myfile.json',  format='pickle')

        It works for `format=joblib` as well,

        .. code:: python

            from pycarol import Carol, Storage
            login = Carol()
            stg = Storage(login)
            my_dict = stg.load(name='myfile.json',  format='joblib')

        Loading a pandas DataFrame

        .. code:: python

            import pandas as pd
            from pycarol import Carol, Storage

            login = Carol()
            stg = Storage(login)

            df = stg.load(name='myfile.parquet', parquet=True)


        """
        return self.backend.load(name=name, format=format, parquet=parquet, cache=cache, storage_space=storage_space,
                                 columns=columns, chunk_size=chunk_size)

    def files_storage_list(self, prefix='pipeline/',  print_paths=False):
        """
        It will return all files in Carol data Storage (CDS).

        Args:

            prefix: `str`, default `pipeline/`
                prefix of the folder to filter the output.
            print_paths: `bool`, default `False`
                Print the tree structure of the files in CDS

        Returns: list of files paths.

        """

        return self.backend.files_storage_list(prefix=prefix, print_paths=print_paths)

    def exists(self, name):
        """

        Check if files exists in Carol Storage.

        Args:

            name: `str`
                Filename

        Returns: `bool`

        """
        return self.backend.exists(name)

    def delete(self, name):
        """

        Delete a file in Carol Storage.

        Args:

            name: `str`
                Filename to be deleted.

        Returns:

        """
        self.backend.delete(name)

    def build_url_parquet_golden(self, dm_name):
        return self.backend.build_url_parquet_golden(dm_name)

    def build_url_parquet_golden_cds(self, dm_name):
        return self.backend.build_url_dask_parquet_golden_cds(dm_name)

    def build_url_parquet_staging(self, staging_name, connector_id):
        return self.backend.build_url_parquet_staging(staging_name, connector_id)

    def build_url_parquet_staging_cds(self, staging_name, connector_id):
        return self.backend.build_url_dask_parquet_staging_cds(staging_name, connector_id)

    def build_url_parquet_staging_master(self, staging_name, connector_id):
        return self.backend.build_url_parquet_staging_master(staging_name, connector_id)

    def build_url_parquet_staging_rejected(self, staging_name, connector_id):
        return self.backend.build_url_parquet_staging_rejected(staging_name, connector_id)

    def build_url_parquet_golden_rejected_cds(self, dm_name):
        return self.backend.build_url_parquet_golden_rejected_cds(dm_name)

    def get_dask_options(self):
        return self.backend.get_dask_options()

    def get_golden_file_paths(self, dm_name):
        return self.backend.get_golden_file_paths(dm_name)

    def get_view_file_paths(self, view_name):
        return self.backend.get_view_file_paths(view_name)

    def get_staging_file_paths(self, staging_name, connector_id):
        return self.backend.get_staging_file_paths(staging_name, connector_id)

    def get_golden_rejected_cds_file_paths(self, dm_name, file_pattern=None):
        return self.backend.get_golden_rejected_cds_file_paths(dm_name, file_pattern=file_pattern)

    def get_staging_cds_file_paths(self, staging_name, connector_id, file_pattern=None):
        return self.backend.get_staging_cds_file_paths(staging_name, connector_id, file_pattern=file_pattern)

    def get_golden_cds_file_paths(self, dm_name, file_pattern=None):
        return self.backend.get_golden_cds_file_paths(dm_name, file_pattern=file_pattern)

    def get_view_cds_file_paths(self, dm_name):
        return self.backend.get_view_cds_file_paths(dm_name)



import redis
import os

class RedisTarget:
    """ Save temporary data in Redis.
        When Redis environment variables are undefined, it will save the data in Memory.

    Usage:
        from pycarol.carol import *
        from pycarol.auth.PwdAuth import *
        from pycarol.redis_target import RedisTarget

        carol = Carol('rui', 'rui', auth=PwdAuth('pedro.buzzi@totvs.com.br', 'Adsl0223'))

        m1 = RedisTarget(carol=carol)
        m1.save('key1', 'value 1')
        print (m1.load('key1'))
    """

    def __init__(self, carol=None, redis_host=None, redis_port=None, redis_password=None):
        if not redis_host:
            redis_host = os.environ.get('REDIS_SERVICE_HOST')
            redis_port = os.environ.get('REDIS_SERVICE_PORT')
            redis_password = os.environ.get('REDIS_SERVICE_PASSWORD', '')

        if redis_host and redis_port:
            self.target = _Redis(carol.domain, carol.app_name, redis_host, redis_port, redis_password)
        else:
            self.target = _Memory()

    def exists(self, name):
        return self.target.exists(name)

    def save(self, name, obj):
        self.target.save(name, obj)

    def load(self, name):
        return self.target.load(name)

    def remove(self, name):
        self.target.remove(name)


class _Redis:
    def __init__(self, domain, app_name, redis_host, redis_port, redis_password):
        self.r = redis.Redis(host=redis_host, port=redis_port, password=redis_password)
        self.hash = '{}/{}'.format (domain, app_name)

    def exists(self, key):
        return self.r.hget(self.hash, key) is not None

    def save(self, key, obj):
        self.r.hset(self.hash, key, obj)

    def load(self, key):
        return self.r.hget(self.hash, key)

    def remove(self, key):
        self.r.hdel(self.hash, key)


class _Memory:
    def __init__(self):
        self.mem_data = {}

    def exists(self, key):
        return key in self.mem_data

    def save(self, key, obj):
        self.mem_data[key] = obj

    def load(self, key):
        return self.mem_data[key]

    def remove(self, key):
        self.mem_data.pop(key, None)



import luigi
import copy
from pycarol.pipeline import Task
from pycarol.pipeline.utils import (
    build_dag,
    breadth_first_search,
    get_reverse_dag,
    find_root_in_dag,
)

from luigi.task import flatten
from collections import defaultdict


class Pipe(object):
    """
    This class should be used to compose a pipeline given a list of tasks. It
    contains many methods to interact with the pipeline as a whole.
    """
    def __init__(self, tasks: list, params:dict):
        assert isinstance(tasks,list)
        assert params is None or isinstance(params,dict)
        assert _tasks_are_class(tasks)
        if _tasks_are_instance(tasks):
            raise NotImplementedError(f"Currently, we cannot create Pipe objects from instantiated tasks.")

        self.params = copy.deepcopy(params)
        self.top_nodes = tasks # top nodes are root nodes
        #self.dag = _get_dag_from_task(tasks)
        self.top_nodes = [t(**self.params) for t in self.top_nodes]
        self.dag = _get_dag_(defaultdict(set), self.top_nodes)
        self.dag = {i: list(j) for i,j in self.dag.items()}
        #self.dag = _get_instances_from_classes(self.dag, self.params)
        self.rev_dag = get_reverse_dag(self.dag)
        self.leaf_nodes = find_root_in_dag(self.rev_dag) #  leaf nodes are root nodes of rev dag
        self.all_tasks = [k for k in self.dag]
        

    def remove_all(self):
        """Remove all targets related to this pipeline."""
        for t in self.all_tasks:
            try:
                t.remove()
            except FileNotFoundError:
                pass

    def remove_upstream(self, tasks:list):
        """Remove all targets in this pipeline that depend on the given tasks."""
        assert isinstance(tasks,list)
        traverse_dag_generator = breadth_first_search( self.rev_dag,tasks)
        for task_list in traverse_dag_generator:
            for t in task_list:
                try:
                    t.remove()
                except FileNotFoundError:
                    pass
    
    def remove_orphans(self):
        """Remove all targets for which respective downstream targets are not complete"""

        downstream_complete_dict = {}
        _downstream_complete(self.dag, self.top_nodes, downstream_complete_dict)

        for t, is_downstream_complete in downstream_complete_dict.items():
            if  not is_downstream_complete:
                try:
                    t.remove()
                except FileNotFoundError:
                    pass

    def remove_obsolete(self):
        """Remove all targets whose hash_versions do not match to current version"""
        return

    def update_all_complete_status(self):
        """ Updates a dictionary whose keys are task objects and values are
        True if target was found."""
        from pycarol import Carol, Storage
        st = Storage(Carol())
        target_files = st.files_storage_list(prefix='pipeline/')
        self.all_complete_status = {t: t.output().path in target_files for t in
                self.all_tasks}

    def get_task_complete(self,task):
        """ Returns True if task is complete accordingly to
        self.all_complete_status dictionary.
        When updating many tasks simultaneously, this method is faster the
        standard luigi method."""
        assert task in self.all_tasks, f"Task {task} not found in this pipeline"
        return self.all_complete_status[task]

    def run(self, local_scheduler=True, workers=1, detailed_summary=False):
        """Run the whole pipeline"""
        tasks = [t for t in self.top_nodes]
        return luigi.build(tasks, local_scheduler=local_scheduler,
                           workers=workers, detailed_summary=detailed_summary)

    def get_dag(self):
        return self.dag

    def get_task_by_id(self,task_id):
        """
        Returns the task object given a string task_id.
        This method should be used to implement run_task command line script,
        which will be used in docker tasks and related architectures.
        """
        for t in self.all_tasks:
            if task_id == t.task_id:
                return t
        else:
            raise KeyError(f"{task_id} not found in this pipeline.")

    def get_matching_tasks(self,task):
        """
        Given a non instantiated task, retrieves all instantiated tasks in this pipeline.
        It is used in manual assert in notebook tasks.
        """
        matching_tasks = [isinstance(t, task) for t in self.all_tasks]
        matching_tasks = [t for i, t in enumerate(self.all_tasks) if
                          matching_tasks[i]]
        return matching_tasks


    def assert_task_is_unique(self,task):
        """
        If more than one instance of a class task exist in this pipeline,
        the use of task notebook will be tricky. This assert should be used
        in notebook tasks to warn the user about this situation.
        """
        matching_tasks = self.get_matching_tasks(task)
        assert len(matching_tasks) > 0, \
            f"task {task} not found in this pipeline"
        assert len(matching_tasks) ==1, \
            f"task {task} found multiple times in this pipeline"
        return

    def get_task_instance(self,task):
        """
        This is the method which allows task notebook feature. Given a task
        class it returns corresponding task object in pipeline
        """
        self.assert_task_is_unique(task)
        matching_tasks = self.get_matching_tasks(task)
        return matching_tasks[0]



### Auxiliary functions ###


def _get_dag_(dag, tasks) -> dict:
    """
    Compute the DAQ of a pipeline.


    Args:
        dag: `collections.defaultdict(set)`
            A defaultdict of a set. This will hold the dict with the result {task : requirements_set}
        tasks: `list`
            List of taks to find the dependencies.

    Returns: collections.defaultdict(set)
        the daq as a collections.defaultdict(set)

    """
    for task in tasks:
        if task in dag:
            continue
        #flatten handles dicts and lists.
        reqs = flatten(task.requires())
        dag[task].update(set(reqs))
        if len(reqs) > 0:
            dag = _get_dag_(dag, reqs)
        else:
            continue

    return dag


def _luigi_get_sons(task) -> list:
    """
    Returns a list of required tasks. This is used in build_dag
    Args:
        task: luigi Task

    Returns:
        l: list of luigi Task

    """
    if isinstance(task,tuple):
        # this case happens when we use local params in tasks require
        task, params = task
    return task.requires_list


def _get_dag_from_task(task:list) -> dict:
    """
    Wrapper around generic build_dag.
    Args:
        task: list of proper luigi tasks

    Returns:
        dag: dict encoding a DAG

    """
    dag = build_dag(task, _luigi_get_sons)
    return dag

def _get_instances_from_classes(dag:dict, params:dict):
    """Returns a dag of task instances, given a dag of task classes and pipeline params."""

    dag = {i: list(j) for i,j in dag.items()}
    instances_dag = {}
    for k,v in dag.items():
        task_params = params
        if isinstance(k,tuple):
            k,local_params = k
            task_params.update(local_params)
        task_list=[]
        for t in v:
            task_params = params
            if isinstance(t,tuple):
                t,local_params = t
                task_params.update(local_params)
            task_list.append(t(**task_params))
        instances_dag[k(**params)] = task_list
    return instances_dag

def _downstream_complete(dag, top_nodes, downstream_complete_dict):
    """Recursively traverses dag starting from top_nodes to update downstream_complet_dict"""
    #TODO: reimplement using breadth_first_search
    for task in top_nodes:
        if task in downstream_complete_dict:
            continue
        sons = dag[task]
        if sons: # recursion step
            downstream_complete_dict[task] = task.complete() and \
                all([
                    _downstream_complete(dag, [t], downstream_complete_dict)
                    for t in sons
                    ])
        else: # stop recursion step
            downstream_complete_dict[task] = task.complete()

def _tasks_are_class(tasks):
    for t in tasks:
        if not issubclass(t,Task):
            return False
    else:
        return True

def _tasks_are_instance(tasks):
    for t in tasks:
        if not isinstance(t,Task):
            return False
    else:
        return True






"""

Allow to copy DataModels, NamedQueries, Connectors, Staging and mappingns from one tenant to another.

"""

from ..data_models.data_models import DataModel, CreateDataModel
from ..named_query import NamedQuery
from ..connectors import Connectors
from ..staging import Staging
from collections import defaultdict


class CloneTenant(object):
    """

    Clone entities from one tenant to a second tenant.

    Args:

        carol_from: `pycarol.Carol`
            `pycarol.Carol` instance to copy from
        carol_to: `pycarol.Carol`
            `pycarol.Carol` instance to copy to

    """

    def __init__(self, carol_from, carol_to):

        self.carol_from = carol_from
        self.carol_to = carol_to

    def copy_named_query(self, named_query_list=None, copy_all=False, overwrite=False):

        """

        Copy named queries.

        Args:
            named_query_list:
            copy_all:
            overwrite:

        """

        named = NamedQuery(self.carol_from)
        all_named = named.get_all()

        if copy_all:
            named_query_list = list(all_named.values())
        else:
            assert named_query_list is not None
            if isinstance(named_query_list, str):
                named_query_list = [named_query_list]
            else:
                assert isinstance(named_query_list, list)

        not_exist = dict()
        not_exist['None'] = []
        name_to_send = []
        for name in named_query_list:
            if all_named.get(name) is not None:
                name_to_send.append(all_named.get(name))
            else:
                not_exist['None'].append(name)

        named_to_send = NamedQuery(self.carol_to)
        named_to_send.create_named_query(name_to_send, overwrite=overwrite)

        if not not_exist['None'] == []:
            print('The following named queries in the list do not exist: {}'.format(not_exist['None']))

        return self

    def copy_data_models(self, dm_list=None, publish=True, overwrite=False, copy_all=False):

        """
        Copy DataModels

        Args:

            dm_list: `list` default `None`
                List of DataModels to copy
            publish: `bool` default `True`
                Publish copied DataModels
            overwrite: `bool` default `False`
                Overwrite if already exists.
            copy_all: `bool` default `False`
                Copy all DataModels. If no list passes, use this to copy all DataModels.

        """

        DMsTenant = DataModel(self.carol_from)

        if copy_all:
            dm_list = list(DMsTenant.get_all().template_dict.keys())
        else:
            assert dm_list is not None
            if isinstance(dm_list, str):
                dm_list = [dm_list]
            else:
                assert isinstance(dm_list, list)

        dm_to_copy = {}
        snapshot_dict = {}
        dm_to_create = CreateDataModel(self.carol_to)

        for dm_name in dm_list:
            DMsTenant.get_by_name(dm_name)
            current_template = DMsTenant.entity_template_.get(dm_name)
            dm_to_copy.update({dm_name: {'mdmId': current_template['mdmId'],
                                         'mdmEntitySpace': current_template['mdmEntitySpace']}})

            DMsTenant.get_snapshot(dm_id=current_template['mdmId'], entity_space=current_template['mdmEntitySpace'])
            current_snap = DMsTenant.snapshot_
            snapshot_dict.update({dm_name: current_snap[dm_name]})
            dm_to_create.from_snapshot(current_snap[dm_name], publish=publish, overwrite=overwrite)

        return self

    def copy_connectors(self, conectors_map, map_type='name', overwrite_connector=False, add_to_connector=True,
                        change_name_dict=None, copy_mapping=True, overwrite_schema=False):
        """
        Copy Connectors, Stagings and mappings.

        Args:

            conectors_map: `dict`
                dictionary mapping the connector and stagings to be copied.
                If `map_type = name`
                    dictionary of {connector_name_1 : [staging_name_1, staging_name_2, staging_name_3 ...],
                                   connector_name_2 : [staging_name_1, staging_name_2, staging_name_3 ...] }
                If `map_type = connector_id`
                    dictionary of {connector_id_1 : [staging_name_1, staging_name_2, staging_name_3 ...],
                                   connector_id_2 : [staging_name_1, staging_name_2, staging_name_3 ...] }

            map_type: 'str' default `name`
                Type of mapping. Possible values: `name` and `connector_id`
            overwrite_connector: `bool` default `False`
                Overwrite connector if it already exists. It will deleate the connector and all stagings in this connector:
            add_to_connector: `bool` default `False`
                If the connector already exists, it will add to the connector.
            change_name_dict: `dict`
                If one wants to change the connector name, this dict is the mapping from the old to the new name.
            copy_mapping: `bool` default `True`
                Copy staging mappings.
            overwrite_schema: `bool` default `False`
                overwrite if the mapping already exists.

        """

        if map_type == 'connector_id':
            map_type = 'mdmId'
        elif map_type == 'name':
            map_type = 'mdmName'
        else:
            raise ('values should be connector_id or name')

        conn_id = {}
        conn = Connectors(self.carol_from)
        conn_to_create = conn.get_all(include_mappings=True)

        stag = Staging(self.carol_from)
        self.stag_mapp_to_use = defaultdict(list)

        for connector, staging in conectors_map.items():
            # for connector in conn_to_create:

            if isinstance(staging, str):
                staging = [staging]

            for list_conn in conn_to_create:
                if list_conn[map_type] == connector:
                    connector = list_conn
                    break
            else:
                raise ValueError('{} does not exist in the tenant'.format(connector))

            current_connector = connector['mdmId']
            conn.stats(connector_id=current_connector)

            if change_name_dict is not None:
                connector_name = change_name_dict.get(connector.get('mdmName', None)).get('name')
                connector_label = change_name_dict.get(connector.get('mdmName', None)).get('label')
                if connector_label is None:
                    connector_label = connector_name
            else:
                connector_name = connector.get('mdmName', None)
                connector_label = connector.get('mdmLabel', None)
                if connector_label:
                    connector_label = connector_label['en-US']
                else:
                    connector_label = None
            group_name = connector.get('mdmGroupName', None)

            conn_to = Connectors(self.carol_to)

            if add_to_connector:
                _con_id = conn_to.get_by_name(connector_name, errors='ignore').get('mdmId')
                if _con_id is None:
                    _con_id = conn_to.create(name=connector_name, label=connector_label,
                                             group_name=group_name, overwrite=overwrite_connector)
            else:
                _con_id = conn_to.create(name=connector_name, label=connector_label,
                                         group_name=group_name, overwrite=overwrite_connector)

            conn_id.update({connector_name: _con_id})

            for schema_name in staging:

                aux_schema = stag.get_schema(staging_name=schema_name, connector_id=connector.get('mdmId'))
                aux_schema.pop('mdmTenantId')

                aux_schema.pop('mdmId')
                aux_schema.pop('mdmCreated')
                aux_schema.pop('mdmLastUpdated')

                stg_to = Staging(self.carol_to)
                stg_to.send_schema(schema=aux_schema, connector_id=conn_id.get(connector_name),
                                   overwrite=overwrite_schema)

                # TODO mappings should be copied after copied all stagings.
                # Need t0 find how to copy ETLs.
                if copy_mapping:

                    mapping_fields = connector.get('mdmEntityMappings', None).get(schema_name)
                    if mapping_fields is not None:
                        mapping_fields.pop('mdmTenantId')
                        mapping_id = mapping_fields.pop('mdmId')
                        entity_space = mapping_fields.get('mdmEntitySpace')
                        mapping_fields.pop('mdmCreated')
                        mapping_fields.pop('mdmLastUpdated')
                        connector_id = mapping_fields.pop('mdmConnectorId')

                        mappings_to_get = stag.get_mapping_snapshot(connector_id=connector_id, mapping_id=mapping_id,
                                                                    entity_space=entity_space)
                        _, aux_map = mappings_to_get.popitem()
                        stg_to.mapping_from_snapshot(mapping_snapshot=aux_map, connector_id=conn_id.get(connector_name),
                                                     overwrite=overwrite_schema)
                        self.stag_mapp_to_use[connector_name].append({"schema": aux_schema, "mapping": aux_map})
                    else:
                        self.stag_mapp_to_use[connector_name].append({"schema": aux_schema})
                else:
                    self.stag_mapp_to_use[connector_name].append({"schema": aux_schema})



from .ApiKeyAuth_cloner import ApiKeyAuthCloner


class ApiKeyAuth:
    """

    Args:
        api_key: `str`
            Api Key

    """

    def __init__(self, api_key):
        self.api_key = api_key
        self.carol = None
        self.connector_id = None
        
    def set_connector_id(self, connector_id):
        self.connector_id = connector_id
    
    def cloner(self):
        return ApiKeyAuthCloner(self)

    def login(self, carol):
        self.carol = carol
        pass

    def authenticate_request(self, headers):
        headers['x-auth-key'] = self.api_key
        headers['x-auth-connectorid'] = self.connector_id

    def switch_context(self, env_id):
        """
        Switch context to an environment within the same organization.

        Args:

            env_id: environment id to switch context to.

        Returns:

            None

        """

        path = f'v2/oauth2/switchTenantContextWithApiKey/{env_id}'
        resp = self.carol.call_api(method='POST', path=path)

        self.api_key = resp['apiKey']
        self.connector_id = resp['connectorId']

    def switch_org_context(self, org_id):
        """
        Go to the organization context or switch organization.

        Args:
            org_id: organization id.

        Returns:
            None

        """

        path = f'v2/oauth2/switchOrgContextWithApiKey/{org_id}'
        resp = self.carol.call_api(method='POST', path=path)

        self.api_key = resp['apiKey']
        self.connector_id = resp['connectorId']





""" PyCarol - Connecting Carol to Python

"""

import os
import tempfile

__version__ = '2.45.2'

__TEMP_STORAGE__ = os.path.join(tempfile.gettempdir(), 'carolina/cache')

__CONNECTOR_PYCAROL__ = 'f9953f6645f449baaccd16ab462f9b64'

_CAROL_METADATA_STAGING = ['mdmCounterForEntity', 'mdmId', 'mdmLastUpdated', 'mdmCreated',
                           'mdmConnectorId', 'mdmDeleted']

_CAROL_METADATA_GOLDEN = ['mdmCounterForEntity', 'mdmStagingCounter', 'mdmId', 'mdmCreated', 'mdmLastUpdated',
                          'mdmTenantId',
                          'mdmEntityType', 'mdmSourceEntityNames', 'mdmCrosswalk', 'mdmStagingRecord',
                          'mdmApplicationIdMasterRecordId',
                          'mdmPreviousIds', 'mdmDeleted']

_CAROL_METADATA_UNTIE_GOLDEN = 'mdmStagingCounter'
_CAROL_METADATA_UNTIE_STAGING = 'mdmCounterForEntity'


_NEEDED_FOR_MERGE = ['mdmCounterForEntity', 'mdmId', 'mdmDeleted']

_REJECTED_DM_COLS = ['mdmConnectorId', 'mdmCreated', 'mdmCreatedUser', 'mdmCrosswalk',
                     'mdmDetailFieldValues', 'mdmEntityTemplateId', 'mdmEntityType',
                     'mdmErrors', 'mdmId', 'mdmLastUpdated', 'mdmMasterFieldAndValues',
                     'mdmStagingApplicationId', 'mdmStagingEntityName', 'mdmStagingRecord',
                     'mdmTenantId', 'mdmUpdatedUser']

from .carol import Carol
from .staging import Staging
from .connectors import Connectors
from .query import Query
from .storage import Storage
from .carolina import Carolina
from .tasks import Tasks
from .data_models import DataModel, DataModelView
from .logger import CarolHandler
from .auth.PwdAuth import PwdAuth
from .auth.ApiKeyAuth import ApiKeyAuth
from .auth.PwdKeyAuth import PwdKeyAuth
from .cds import CDSGolden, CDSStaging
from .apps import Apps
from .subscription import Subscription
from .carol_api import CarolAPI



from .utils.deprecation_msgs import _deprecation_msgs


class Tenant:
    def __init__(self, carol):
        self.carol = carol

    def get_tenant_by_domain(self, domain, auth=True):
        """
        Get tenant information.

        Args:
            domain: `str`
                Tenant name
            auth: `bool` default `True`
                This API ca be called without being logged in. If auth=False it will not use token to call it.

        Returns:
            dict with the information about the tenant.
        """
        return self.carol.call_api('v2/tenants/domain/{}'.format(domain), auth=auth, status_forcelist=[], retries=0)


    def check_exports(self, staging=True, data_model=True):
        """

        @DEPRECATED. This function was removed in pycarol 3.34

        Find all data models or stating tables that have export active.

        :param staging: `bool`, default `True`
            Get status for exports in staging.
        :param data_model: `bool`, default `True`
            Get status for exports in data models.
        :return: `tuple`
            staging stats, data model stats
        """

        _deprecation_msgs("This function was removed from pyCarol")

        return None



from . import data_models


class DataModelBuild:
    """ Automatize the process of building and maintaining a Data Model

        Methods to get, interpret and update information about data model structure, which includes:

            * Data Models
            * Fields
            * Validation Rules
    """

    def __init__(self, carol):
        self.carol = carol

        self.dms = None
        self.dms_id = None

        self.field_functions = None
        self.field_functions_id = None

    def get_all_dm(self):
        """ Set up data models

            Updates:
                self.dms: Dict with Data Model information, following MDM Model
                self.dms_ds: Dict with {data_model_name:data_model_id} for convenience

        :return: self
        """
        dm = data_models.DataModel(self.carol).get_all()
        self.dms = dm.template_data
        self.dms_id = {d['mdmName']: d['mdmId'] for d in self.dms}
        return self

    def get_dm(self, dm_name):
        """ Get a data model by its name

        :param dm_name: Data Model name
        :return: Dict with selected data model information following MDM Model
        """
        if self.dms is None:
            self.get_all_dm()
        for d in self.dms:
            if d['mdmName'] == dm_name:
                return d

    def get_fields(self, dm_name):
        """ Get a fields from a data model by its name

        :param dm_name: Data Model name

        :return: Dict with selected data model fields following MDM Model
        """
        dm = self.get_dm(dm_name)
        return dm['mdmFields']

    def get_fields_names(self, dm_name):
        return [f['mdmName'] for f in self.get_fields(dm_name)]

    def select_field(self, dm_name, field_name):
        """ Get a specific field from a data model

        :param dm_name: Data Model name
        :param field_name: Field name in data model

        :return: Dict with selected data model and field following MDM Model
        """
        fields = self.get_fields(dm_name)
        # TODO: Nested fields
        for f in fields:
            if f['mdmName'] == field_name:
                return f

    def _get_all_field_functions(self):
        """ Set up all possible field functions

            Updates:
                self.field_functions:
                self.field_functions_id:

        :return: None
        """
        get_url = 'v1/fieldFunctions'
        self.field_functions = {f['mdmId']: f for f in self.carol.call_api(get_url, params=dict(pageSize=-1))['hits']}
        self.field_functions_id = {f['mdmName']: f['mdmId'] for f in self.field_functions.values()}

    def get_field_function(self, function_name):
        """ Get a function by its name

        :param function_name: Name of the field function

        :return: Dict of Field Function Id : Field function following MDM Model
        """
        if self.field_functions is None:
            self._get_all_field_functions()
        return self.field_functions[self.field_functions_id[function_name]]

    def filter_rejection_rule_functions(self):
        """ Get only possible rejection rules functions from field functions

        :return: Dict of Field Function Id : Field function following MDM Model
        """
        self.get_all_field_functions()
        # TODO: Get rejection rules from API
        rejection_rules = ['CONTAINS', 'CONTAINS_ARRAY', 'STARTS_WITH', 'ENDS_WITH', 'EQUALS',
                           'MATCHES', 'IS_EMPTY', 'IS_EMAIL', 'LENGTH_COMPARE', 'LENGTH_RANGE',
                           'IS_NOT_NUMBER_ONLY', 'IS_INVALID_BRAZIL_STATE_TAX_ID',
                           'IS_INVALID_BRAZIL_TAX_ID', 'VALUE_COMPARE', 'LOOKUP', 'LOOKUP_MULTIPLE',
                           'LOOKUP_FOUND']
        return {f['mdmId']: f for f in self.field_functions if f['mdmName'] in self.field_functions.values()}

    def get_dm_rule_ids(self, dm_name):
        """ Get data model's validation rules

        :param dm_name: Data model name
        :return:
        """
        dm = self.get_dm(dm_name)
        return dm['mdmEntityValidationRuleIds']

    def get_dm_rule(self, ruleId, templateId):
        """ Get specific rule from data_model

        :param ruleId: Id of the validation rule
        :param templateId: Id of the template
        :return: Validation Rule following MDM Model
        """
        get_url = f'v1/entities/templates/{templateId}/entityValidationRules/{ruleId}'
        return self.carol.call_api(get_url)

    def create_rule(self, dm_name, field_name, rule_name, params):
        """ Create a rule at Carol
            :param dm_name: data model name
            :param field_name:
            :param rule_name:
            :param params: list with parameters for rule, without the fieldId (will be generated automatically)
        """
        templateId = self.get_dm(dm_name)['mdmId']
        post_url = f'v1/entities/templates/{templateId}/entityValidationRules'
        body = self.generate_rule_body(templateId, dm_name, field_name, rule_name, params)
        rule = self.carol.call_api(post_url, data=body, method='POST')

        post_url = f'v1/entities/templates/{templateId}'
        rule_id = rule['mdmId']
        update_id = {"mdmEntityValidationRuleIds": [[rule_id]]}
        return rule_id, self.carol.call_api(post_url, data=update_id, method='PUT')

    def create_rule_from_uniform_type(self, dm_name, field_name, uniform_type):
        """ Use uniform type to generate a new rule

        :param dm_name: Data Model Name
        :param field_name: Field Name
        :param uniform_type: Uniform Type object

        :return: success
        """
        return self.create_rule(dm_name, field_name, 'MATCHES', [uniform_type.create_regex()])

    def generate_rule_body(self, templateId, dm_name, field_name, rule_name, params):
        """ Generate body for API call to create new Validation Rule

        :param templateId: Template ID
        :param dm_name: Data Model name
        :param field_name: Field Name
        :param rule_name: Rule Name
        :param params: Parameters (list)

        :return:
        """
        field_func = self.get_field_function(rule_name)
        field_id = self.select_field(dm_name, field_name)['mdmId']

        body = dict(
            mdmEntityTemplateId=templateId,
            mdmFieldFunctionId=field_func['mdmId'],
            mdmFieldId=field_id,
            mdmParameterValues=[f"@@{field_id}"] + params
        )
        return body

    def delete_rule(self, dm_name, rule_id):
        # TODO
        pass

    def update_rule(self, dm_name, rule_id, field_name, params):
        # TODO
        pass

    def add_skip_rule(self, dm_name, field_name, rule):
        # TODO
        pass

    def delete_skip_rule(self, dm_name, field_name, rule):
        # TODO
        pass

    def update_skip_rule(self, dm_name, rule_id, update):
        # TODO
        pass


from pycarol.pipeline.task import Task, inherit_list
from pycarol.pipeline.tools import Pipe

from functools import partial
import numpy as np
#TODO: improve pipeline_example
def f1():
    return np.ones([50,1])

def f2():
    return np.ones([1,40])

def f3(x,y,**params):
    return x*y

class T1(Task):
    task_function = f1


class T2(Task):
    task_function = f2


@inherit_list(T1,T2)
class T3(Task):
    task_function = f3

params = {}

pipeline1 = Pipe([T3],params)


import os
import pickle
import gzip
from .. import __TEMP_STORAGE__
from collections import defaultdict
from ..utils.miscellaneous import prettify_path, _attach_path, _FILE_MARKER
from retry import retry
from google.resumable_media import DataCorruption
from google.api_core.exceptions import GatewayTimeout, ServiceUnavailable

CDS_RETRY_LIST = (GatewayTimeout, DataCorruption, ServiceUnavailable)

class StorageGCPCS:
    def __init__(self, carol, carolina):
        import pandas as pd
        self.carol = carol
        self.carolina = carolina
        os.makedirs(__TEMP_STORAGE__, exist_ok=True)

    def _get_app_storage_bucket(self):
        return self.carolina.get_client().bucket(self.carolina.get_bucket_name("app"))

    def save(self, name, obj, format='pickle', parquet=False, cache=True, chunk_size=None):
        """

        Args:

            name: `str`.
                Filename to be used when saving the `obj`
            obj: `obj`
                It depends on the `format` parameter.
            format: `str`
                Possible values:

                    1. `pickle`: It uses `pickle.dump` to save the binary file.
                    2. `joblib`: It uses `joblib.dump` to save the binary file.
                    3. `file`: It saves a local file sending it directly to Carol.

            parquet: `bool` default `False`
                It uses `pandas.DataFrame.to_parquet` to save. `obj` should be a pandas DataFrame
            cache: `bool` default `True`
                Cache the file saved in the temp directory.
            chunk_size: `int` default `None`
                The size of a chunk of data whenever iterating (in bytes).
                This must be a multiple of 256 KB per the API specification.
        """

        remote_file_name = f"{self.carolina.get_path('app', {})}{name}"
        local_file_name = os.path.join(__TEMP_STORAGE__, remote_file_name.replace("/", "-"))

        bucket = self._get_app_storage_bucket()
        blob = bucket.blob(remote_file_name, chunk_size=chunk_size)

        if parquet:
            import pandas as pd
            if not isinstance(obj, pd.DataFrame):
                raise ValueError(f"Object to be saved as parquet must be a "
                                 f"DataFrame. Received a {type(obj)}")
            obj.to_parquet(local_file_name)
        elif format == 'joblib':
            import joblib

            if not cache:
                from io import BytesIO
                with BytesIO() as buffer:
                    joblib.dump(obj, buffer)
                    buffer.seek(0)
                    blob.upload_from_file(buffer)
                return
            else:
                joblib.dump(obj, local_file_name)
        elif format == 'pickle':
            with gzip.open(local_file_name, 'wb') as f:
                pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)
        elif format == 'file':
            local_file_name = obj
        else:
            raise ValueError("Supported formats are pickle, joblib or file")

        blob.upload_from_filename(filename=local_file_name)
        os.utime(local_file_name, None)

    @retry(CDS_RETRY_LIST, tries=5)
    def load(self, name, format='pickle', parquet=False, cache=True, storage_space='app_storage', columns=None,
             chunk_size=None):
        """

        Args:

            name: `str`.
                Filename to be load
            format: `str`
                Possible values:

                    1. `pickle`: It uses `pickle.dump` to save the binary file.
                    2. `joblib`: It uses `joblib.dump` to save the binary file.
                    3. `file`: It saves a local file sending it directly to Carol.

            parquet: `bool` default `False`
                It uses `pandas.DataFrame.to_parquet` to save. `obj` should be a pandas DataFrame
            cache: `bool` default `True`
                Cache the file saved in the temp directory.
            storage_space: `str` default `app_storage`
                Internal Storage space.
                    1. "app_storage": For raw storage.
                    2. "golden": Data Model golden records.
                    3. "staging": Staging records path
                    4. "staging_master": Staging records from Master
                    5. "staging_rejected": Staging records from Rejected
                    6. "view": Data Model View records
                    7. "app": App  bucket
                    8. "golden_cds": CDS golden records
                    9. "staging_cds": Staging Intake.
            columns: `list` default `None`
                Columns to fetch when using `parquet=True`
            chunk_size: `int` default `None`
                The size of a chunk of data whenever iterating (in bytes).
                This must be a multiple of 256 KB per the API specification.

        """

        if storage_space == 'app_storage':
            remote_file_name = f"{self.carolina.get_path('app', {})}{name}"
            bucket = self._get_app_storage_bucket()
        else:
            remote_file_name = name
            bucket = self.carolina.get_client().bucket(self.carolina.get_bucket_name(storage_space))

        local_file_name = os.path.join(__TEMP_STORAGE__, remote_file_name.replace("/", "-"))

        has_cache = cache and os.path.isfile(local_file_name)

        if has_cache:
            localts = os.stat(local_file_name).st_mtime
        else:
            localts = 0

        blob = bucket.blob(remote_file_name, chunk_size=chunk_size)
        if not blob.exists():
            return None

        blob.reload()
        remote_ts = blob.updated.timestamp()

        if not cache and format == 'raw':
            import joblib
            from io import BytesIO
            buffer = BytesIO()
            blob.download_to_file(buffer)
            return buffer

        elif not cache and format == 'joblib':
            import joblib
            from io import BytesIO
            with BytesIO() as buffer:
                blob.download_to_file(buffer)
                return joblib.load(buffer)

        # Local cache is outdated
        if localts < remote_ts:
            blob.download_to_filename(local_file_name)

        if os.path.isfile(local_file_name):
            if parquet:
                import pandas as pd
                return pd.read_parquet(local_file_name, columns=columns)
            elif format == 'joblib':
                import joblib
                return joblib.load(local_file_name)
            elif format == 'pickle':
                with gzip.open(local_file_name, 'rb') as f:
                    return pickle.load(f)
            elif format == 'file':
                return local_file_name
            else:
                raise ValueError("Supported formats are pickle, joblib or file")
        else:
            return None

    def exists(self, name):
        remote_file_name = f"{self.carolina.get_path('app', {})}{name}"

        blob = self._get_app_storage_bucket().blob(remote_file_name)
        return blob.exists()

    def delete(self, name):
        remote_file_name = f"{self.carolina.get_path('app', {})}{name}"

        blob = self._get_app_storage_bucket().blob(remote_file_name)
        if blob.exists():
            blob.delete()

        local_file_name = os.path.join(__TEMP_STORAGE__, remote_file_name.replace("/", "-"))
        if os.path.isfile(local_file_name):
            os.remove(local_file_name)

    def build_url_parquet_golden(self, dm_name):
        path = self.carolina.get_path("golden", {'dm_name': dm_name})
        return f'gcs://{self.carolina.get_bucket_name("golden")}/{path}'

    def build_url_parquet_golden_cds(self, dm_name):
        path = self.carolina.get_path("golden_cds", {'dm_name': dm_name})
        return f'gcs://{self.carolina.get_bucket_name("golden_cds")}/{path}'

    def build_url_dask_parquet_golden_cds(self, dm_name):
        bucket = self.carolina.get_client().bucket(self.carolina.get_bucket_name('golden_cds'))
        path = self.carolina.get_path("golden_cds", {'dm_name': dm_name})
        blobs = bucket.list_blobs(prefix=path, delimiter=None)
        #Dask does not accept iterators.
        return [f'gcs://{file.bucket.name}/{file.name}' for file in blobs]

    def build_url_parquet_staging_cds(self, staging_name, connector_id):
        path = self.carolina.get_path('staging_cds', {'connector_id': connector_id, 'staging_type': staging_name})
        return f'gcs://{self.carolina.get_bucket_name("staging_cds")}/{path}'

    def build_url_dask_parquet_staging_cds(self, staging_name, connector_id):

        bucket = self.carolina.get_client().bucket(self.carolina.get_bucket_name('staging_cds'))
        path = self.carolina.get_path('staging_cds', {'connector_id': connector_id, 'staging_type': staging_name})
        blobs = bucket.list_blobs(prefix=path, delimiter=None)
        #Dask does not accept iterators.
        return [f'gcs://{file.bucket.name}/{file.name}' for file in blobs]

    def build_url_parquet_view(self, view_name):
        path = self.carolina.get_path("view", {'relationship_view_name': view_name})
        return f'gcs://{self.carolina.get_bucket_name("view")}/{path}'

    def build_url_parquet_staging(self, staging_name, connector_id):
        path = self.carolina.get_path("staging", {'connector_id': connector_id, 'staging_type': staging_name})
        bucket_name = self.carolina.get_bucket_name("staging")
        bucket = self.carolina.get_client().bucket(bucket_name)
        blobs = list(bucket.list_blobs(prefix=path))
        if len(blobs) == 0:
            return None
        return f'gcs://{bucket_name}/{path}'

    def build_url_parquet_staging_master(self, staging_name, connector_id):
        path = self.carolina.get_path("staging_master", {'connector_id': connector_id, 'staging_type': staging_name})
        bucket_name = self.carolina.get_bucket_name("staging_master")
        bucket = self.carolina.get_client().bucket(bucket_name)
        blobs = list(bucket.list_blobs(prefix=path))
        if len(blobs) == 0:
            return None
        return f'gcs://{bucket_name}/{path}'

    def build_url_parquet_staging_rejected(self, staging_name, connector_id):
        path = self.carolina.get_path("staging_rejected", {'connector_id': connector_id, 'staging_type': staging_name})
        bucket_name = self.carolina.get_bucket_name("staging_rejected")
        bucket = self.carolina.get_client().bucket(bucket_name)
        blobs = list(bucket.list_blobs(prefix=path))
        if len(blobs) == 0:
            return None
        return f'gcs://{bucket_name}/{path}'

    def build_url_parquet_golden_rejected_cds(self, dm_name):
        path = self.carolina.get_path("golden_rejected", {'dm_name': dm_name})
        return f'gcs://{self.carolina.get_bucket_name("golden_rejected")}/{path}'

    def get_dask_options(self):
        return {'token': self.carolina.token}

    # TODO All of these could be a single function.
    def get_golden_file_paths(self, dm_name):
        bucket = self.carolina.get_client().bucket(self.carolina.get_bucket_name('golden'))
        path = self.carolina.get_path('golden', {'dm_name': dm_name})
        blobs = bucket.list_blobs(prefix=path, delimiter=None)
        return ({'storage_space': 'golden', 'name': i.name} for i in blobs if i.name.endswith('.parquet'))

    def get_golden_cds_file_paths(self, dm_name, file_pattern=None):
        bucket = self.carolina.get_client().bucket(self.carolina.get_bucket_name('golden_cds'))
        path = self.carolina.get_path('golden_cds', {'dm_name': dm_name})
        if file_pattern is not None:
            path = path + file_pattern

        blobs = bucket.list_blobs(prefix=path, delimiter=None)
        return ({'storage_space': 'golden_cds', 'name': i.name} for i in blobs if i.name.endswith('.parquet'))

    def get_view_cds_file_paths(self, dm_name):
        bucket = self.carolina.get_client().bucket(self.carolina.get_bucket_name('view_cds'))
        path = self.carolina.get_path('view_cds', {'dm_name': dm_name})
        blobs = bucket.list_blobs(prefix=path, delimiter=None)
        return ({'storage_space': 'view_cds', 'name': i.name} for i in blobs if i.name.endswith('.parquet'))

    def get_view_file_paths(self, view_name):
        bucket = self.carolina.get_client().bucket(self.carolina.get_bucket_name('view'))
        path = self.carolina.get_path('view', {'relationship_view_name': view_name})
        blobs = bucket.list_blobs(prefix=path, delimiter=None)
        return ({'storage_space': 'view', 'name': i.name} for i in blobs if i.name.endswith('.parquet'))

    def get_staging_cds_file_paths(self, staging_name, connector_id, file_pattern=None):

        bucket = self.carolina.get_client().bucket(self.carolina.get_bucket_name('staging_cds'))
        path = self.carolina.get_path('staging_cds', {'connector_id': connector_id, 'staging_type': staging_name})
        if file_pattern is not None:
            path = path + file_pattern

        blobs = bucket.list_blobs(prefix=path, delimiter=None)
        return ({'storage_space': 'staging_cds', 'name': i.name} for i in blobs if i.name.endswith('.parquet'))

    def get_staging_file_paths(self, staging_name, connector_id):
        bucket_staging = self.carolina.get_client().bucket(self.carolina.get_bucket_name('staging'))
        path_staging = self.carolina.get_path("staging", {'connector_id': connector_id, 'staging_type': staging_name})
        blobs_staging = list(bucket_staging.list_blobs(prefix=path_staging))

        bucket_master = self.carolina.get_client().bucket(self.carolina.get_bucket_name('staging_master'))
        path_master = self.carolina.get_path("staging_master",
                                             {'connector_id': connector_id, 'staging_type': staging_name})
        blobs_master = list(bucket_master.list_blobs(prefix=path_master))

        bucket_rejected = self.carolina.get_client().bucket(self.carolina.get_bucket_name('staging_rejected'))
        path_rejected = self.carolina.get_path("staging_rejected",
                                               {'connector_id': connector_id, 'staging_type': staging_name})
        blobs_rejected = list(bucket_rejected.list_blobs(prefix=path_rejected))

        bs = [{'storage_space': 'staging', 'name': i.name} for i in blobs_staging if i.name.endswith('.parquet')]
        bm = [{'storage_space': 'staging_master', 'name': i.name} for i in blobs_master if i.name.endswith('.parquet')]
        br = [{'storage_space': 'staging_rejected', 'name': i.name} for i in blobs_rejected if
              i.name.endswith('.parquet')]

        ball = bs + bm + br
        return ball

    def get_golden_rejected_cds_file_paths(self, dm_name, file_pattern=None):
        bucket_rejected = self.carolina.get_client().bucket(self.carolina.get_bucket_name('golden_rejected'))
        path_rejected = self.carolina.get_path("golden_rejected", {'dm_name': dm_name, })
        if file_pattern is not None:
            path_rejected = path_rejected + file_pattern

        blobs_rejected = list(bucket_rejected.list_blobs(prefix=path_rejected))

        br = [{'storage_space': 'golden_rejected', 'name': i.name} for i in blobs_rejected if
              i.name.endswith('.json.gz')]
        return br

    def files_storage_list(self, prefix='pipeline/', print_paths=False):
        bucket_staging = self.carolina.get_client().bucket(self.carolina.get_bucket_name('app'))
        path_app = self.carolina.get_path('app', {})

        files = list(bucket_staging.list_blobs(prefix=path_app + prefix))
        files = [i.name.split(path_app)[-1] for i in files]
        if print_paths:
            main_dict = defaultdict(dict, ((_FILE_MARKER, []),))
            for line in files:
                _attach_path(line, main_dict)
            prettify_path(main_dict)

        return files



import json
from collections import defaultdict
from .utils.deprecation_msgs import _deprecation_msgs


class Connectors:
    """
    This class handle all APIs related to Carol connectors

    Args:

        carol: class: pycarol.Carol

    """

    def __init__(self, carol):
        self.carol = carol

    def create(self, name, label=None, group_name="Others", overwrite=False):
        """
        Create a connector

        Args:

            name: 'str'
                Connector name
            label: 'str'
                Connector label in UI
            group_name: `str` default "Others"
                Connector group name in UI
            overwrite: `bool` default `False`
                Overwrite if already exists.
        """

        if label is None:
            label = name

        resp = self.carol.call_api('v1/connectors', data={
            'mdmName': name,
            'mdmGroupName': group_name,
            'mdmLabel': {"en-US": label}
        }, errors='ignore')
        if resp.get('mdmId') is not None:
            return resp.get('mdmId')
        if ('already exists' in resp.get('errorMessage', [])):
            if overwrite:
                self.delete_by_name(name)
                return self.create(name, label, group_name, False)
            else:
                return self.get_by_name(name)['mdmId']

        else:
            raise Exception(resp)

    def get_by_name(self, name, errors='raise'):
        """
        Get connector information using the connector name

        Args:

            name: 'str'
                Connector Name
            errors: {‘ignore’, ‘raise’}, default ‘raise’
                If ‘raise’, then invalid request will raise an exception If ‘ignore’,
                then invalid request will return the request response


        Returns: `dict`
            connector information.

        """

        resp = self.carol.call_api(f'v1/connectors/name/{name}', errors=errors)
        return resp

    def get_by_id(self, id, errors='raise'):
        """
        Get connector information using the connector name

        Args:

            id: 'str'
                Connector ID
            errors: {‘ignore’, ‘raise’}, default ‘raise’
                If ‘raise’, then invalid request will raise an exception If ‘ignore’,
                then invalid request will return the request response


        Returns: `dict`
            connector information.

        """

        resp = self.carol.call_api(f'v1/connectors/{id}', errors=errors)
        return resp

    def delete_by_name(self, name, force_deletion=True):
        """
        Delete connector by name

        Args:

            name: `str`
                Connector name
            force_deletion: `bool` default `True`
                Force the deletion

        Returns: None

        """

        mdm_id = self.get_by_name(name)['mdmId']
        self.delete_by_id(mdm_id, force_deletion)

    def delete_by_id(self, connector_id=None, mdm_id=None, force_deletion=True):
        """
        Delete Connector by ID

        Args:

            connector_id: `str``
                Connector ID
            mdm_id: `str``
                Connector ID
            force_deletion: `bool` default `True`
                Force the deletion

        Returns: None

        """

        if connector_id is None and mdm_id is not None:
            _deprecation_msgs(
                "mdm_id is deprecated and will be removed, use connector_id")

        connector_id = connector_id if connector_id else mdm_id

        if connector_id is None:
            raise ValueError('Connector Id must be set')

        self.carol.call_api(
            f'v1/connectors/{connector_id}?forceDeletion={force_deletion}', method='DELETE')

    def get_all(self, offset=0, page_size=-1, sort_order='ASC', sort_by=None, include_connectors=False,
                include_mappings=False, include_consumption=False, print_status=True, save_results=False,
                filename='connectors.json'):
        """
        Get all connectors.

        Args:

            offset: `int`, default `0`
                Offset for the response.
            page_size: `int`, default `1000`
                Number of records in each call.
            sort_by: `str` default `None`
                Field to sort by
            sort_order: `str`, default `ASC`
                Sort Order. Possible values "ASC" and "DESC"
            include_connectors: `bool` default `False`
                Include connector information
            include_mappings: `bool` default `False`
                Include mapping information
            include_consumption: `bool` default `False`
                Include consumption information
            print_status: `bool` default `True`
                Print status of the request.
            save_results: `bool` default `False`
                If save json with the results.
            filename: `str` default `None`
                Filename to save

        Returns: `dict`
            Connector information

        """

        params = {"offset": offset, "pageSize": str(page_size), "sortOrder": sort_order,
                  "includeMappings": include_mappings, "includeConsumption": include_consumption,
                  "includeConnectors": include_connectors}

        if sort_by is not None:
            params['sortBy'] = sort_by

        connectors = []
        set_param = True
        count = offset
        total_hits = float("inf")
        if save_results:
            file = open(filename, 'w', encoding='utf8')
        while count < total_hits:
            conn = self.carol.call_api('v1/connectors', params=params)
            count += conn['count']
            if set_param:
                total_hits = conn["totalHits"]
                set_param = False
            conn = conn['hits']

            if not conn:
                break

            connectors.extend(conn)
            params['offset'] = count
            if print_status:
                print('{}/{}'.format(count, total_hits), end='\r')
            if save_results:
                file.write(json.dumps(conn, ensure_ascii=False))
                file.write('\n')
                file.flush()
        if save_results:
            file.close()
        return connectors

    def stats(self, connector_id=None, connector_name=None, all_connectors=False):
        """
        Get connector stats

        Args:

            connector_id: `str`, default `None`
                Connector Id
            connector_name: `str`, default `None`
                Connectot name
            all_connectors: `bool` default `False`
                Get from all connectors.

        Returns: `dict`
            Dict with the status of the connectors.

        """

        if all_connectors:
            response = self.carol.call_api('v1/connectors/stats/all')
        else:
            if connector_name:
                connector_id = self.get_by_name(connector_name)['mdmId']
            else:
                assert connector_id

            response = self.carol.call_api(
                'v1/connectors/{}/stats'.format(connector_id))

        self._conn_stats = response['aggs']
        return {key: list(value['stagingEntityStats'].keys()) for key, value in self._conn_stats.items()}

    def staging_to_connectors_map(self):
        """
        Create a dictionary where the mapping of connectors and stagings.

        Returns: `dict`
            Dict

        """

        d = defaultdict(list)
        connectors = self.get_all(print_status=False)
        for connector in connectors:
            current_connector = connector['mdmId']
            conn_stats = self.stats(current_connector)
            for i in conn_stats[current_connector]:
                d[i].append(current_connector)

        return d

    def find_by_staging(self, staging_name=None):
        """
        Find connector given a staging table

        Args:

            staging_name: `str` default `None`
                Staging table name

        Returns: `dict`
            Connector information

        """

        d = self.staging_to_connectors_map()

        if staging_name:
            conn = d.get(staging_name, None)
            if conn is None:
                raise ValueError(
                    'There is no staging named {}'.format(staging_name))

            elif len(conn) > 1:
                print('More than one connector with the staging {}'.format(
                    staging_name))
                return conn

    def get_dm_mappings(self, connector_id=None, connector_name=None, staging_name=None,
                        dm_id=None, dm_name=None, reverse_mapping=False, offset=0, page_size=1000, sort_by=None,
                        sort_order='ASC', all_connectors=False):
        """
        Get data models mappings information.

        Args:

            connector_id: `str`, default `None`
                Connector Id
            connector_name: `str`, default `None`
                Connectot name
            staging_name: `str`, default `None`
                Staging name
            dm_id: `str`, default `None`
                Data model Id
            dm_name: `str`, default `None`
                Data model name
            reverse_mapping: `bool`, default `False`
                If to return the reverse mapping.
            offset: `int`, default `0`
                Offset for the response.
            page_size: `int`, default `1000`
                Number of records in each call.
            sort_by: `str` default `None`
                Field to sort by
            sort_order: `str`, default `ASC`
                Sort Order. Possible values "ASC" and "DESC"
            all_connectors: `bool`, default `False`
                It will return all the mapping for all connectors/stagings

        Returns: `dict`
            Mapping json definition

        """

        if all_connectors:
            payload = {
                "offset": offset,
                "sortBy": sort_by,
                "pageSize": page_size,
                "sortOrder": sort_order
            }
            url = "v1/connectors/mappings/all"

        else:
            if connector_id is None and connector_name is None:
                raise ValueError(
                    'Either connector_id or connector_name must be set if `all_connectors=False`. ')
            connector_id = connector_id if connector_id else self.get_by_name(connector_name)[
                'mdmId']

            if dm_name is not None:
                url_dm = f"v1/entities/templates/name/{dm_name}"
                dm_id = self.carol.call_api(url_dm, method='GET')['mdmId']

            payload = {
                "reverseMapping": reverse_mapping,
                "entityId": dm_id,
                "stagingType": staging_name,
                "offset": offset,
                "sortBy": sort_by,
                "pageSize": page_size,
                "sortOrder": sort_order
            }

            url = f"v1/connectors/{connector_id}/entityMappings"

        set_param = True
        to_get = float('inf')
        count = 0
        self.resp = []
        while count < to_get:

            resp = self.carol.call_api(url, method='GET', params=payload)
            if set_param:
                self.total_hits = resp["totalHits"]
                to_get = resp["totalHits"]
                set_param = False

            count += resp['count']
            query = resp['hits']
            if not query:
                break

            self.resp.extend(query)
            payload['offset'] = count

        return self.resp

    def get_entity_mappings(
            self, connector_name=None, connector_id=None,
            reverse_mapping=False, staging_name=None,
            offset=0, page_size=1000, sort_order='ASC',
            sort_by=None, print_status=False, errors='raise'
    ):
        """
        Get all Entity Mappings.

        Args:
            connector_name: `str`, `str`, default `None`
                Connector Name
            connector_id: `str`, `str`, default `None`
                Connector ID
            reverse_mapping: `bool` default `False`
                When using with consumer.False if you don't know what consumer is.
            staging_name: str` default None
                Name of the staging to find the mapping. Returns 404  if there is not mapping..
            offset: `int`, default 0
                Offset for pagination. Only used when `scrollable=False`
            page_size: `int`, default 100
                Number of records downloaded in each pagination. The maximum value is 1000
            sort_order: `str`, default 'ASC'
                Sort ascending ('ASC') vs. descending ('DESC').
            sort_by: `str`,  default `None`
                Name to sort by.
            print_status: `bool`, default `False`
                Print the number of records in each interaction.
            errors: {‘ignore’, ‘raise’}, default ‘raise’
                If ‘raise’, then invalid request will raise an exception If ‘ignore’,
                then invalid request will return the request response.

        Returns: list of dict
            List of dict with mappings,
        """

        connector_id = connector_id if connector_id else self.get_by_name(connector_name)[
            'mdmId']

        template_data = []
        count = offset
        query_params = {
            "offset": offset, "pageSize": str(page_size),
            "sortOrder": sort_order,
            "sortBy": sort_by,
            'stagingType': staging_name,
            'reverseMapping': reverse_mapping,
        }

        set_param = True
        total_hits = float("inf")
        while count < total_hits:
            query = self.carol.call_api(path=f'v1/connectors/{connector_id}/entityMappings', method="GET",
                                        params=query_params, errors=errors)

            if query.get('hits') is None:
                if len(template_data) == 0:
                    # when errors==ignore it will return the error msg.
                    return query
                else:
                    return template_data

            if query['count'] == 0:
                print('There are no more results.')
                print(f'Expecting {total_hits}, response = {count}')
                break
            count += query['count']
            if set_param:
                total_hits = query["totalHits"]
                set_param = False

            query = query['hits']
            template_data.extend(query)

            query_params['offset'] = count
            if print_status:
                print(f'{count}/{total_hits}', end='\r')

        return template_data

    def _play_pause_mapping(self, kind, entity_mapping_id=None, staging_name=None,
                            connector_name=None, connector_id=None,
                            reverse_mapping=False,
                            process_cds=True, ):

        connector_id = connector_id if connector_id else self.get_by_name(connector_name)[
            'mdmId']

        if entity_mapping_id is None:
            if staging_name is None:
                raise ValueError(
                    "Either staging_name or entity_mapping_id must be set.")
            entity_mappings = self.get_entity_mappings(
                connector_id=connector_id, staging_name=staging_name)
            entity_mappings = [i['mdmId'] for i in entity_mappings]
        else:
            if isinstance(entity_mapping_id, str):
                entity_mappings = [entity_mapping_id]
            elif isinstance(entity_mapping_id, list):
                entity_mappings = entity_mapping_id
            else:
                raise ValueError(
                    'entity_mapping_id must be string of list of string.')

        responses = {}
        for _mapping in entity_mappings:
            entity_mapping_id = _mapping

            params = {
                'reverseMapping': reverse_mapping,
                'processCds': process_cds,
            }

            resp = self.carol.call_api(path=f'v1/connectors/{connector_id}/entityMappings/{entity_mapping_id}/{kind}',
                                       method="POST", params=params, )
            responses[entity_mapping_id] = resp

        return responses

    def play_mapping(
            self, entity_mapping_id=None, staging_name=None,
            connector_name=None, connector_id=None,
            reverse_mapping=False,
            process_cds=True,
    ):
        """
        Start mapping.

        Args:
            entity_mapping_id: `str` or list of strings
                Mapping ids to be resumed.
            staging_name:
                Staging name for starting the mapping
            connector_name: `str`, `str`, default `None`
                Connector Name
            connector_id: `str`, `str`, default `None`
                Connector ID
            reverse_mapping: `bool` default `False`
                When using with consumer.False if you don't know what consumer is.
            process_cds: `bool` default `True`
                Process pending records after play.

        Returns: dict
         Dictionary with the response of all mappings played.

        """

        responses = self._play_pause_mapping(
            kind='play', entity_mapping_id=entity_mapping_id, staging_name=staging_name,
            connector_name=connector_name, connector_id=connector_id,
            reverse_mapping=reverse_mapping,
            process_cds=process_cds,
        )

        return responses

    def pause_mapping(
            self, entity_mapping_id=None, staging_name=None,
            connector_name=None, connector_id=None,
            reverse_mapping=False,
    ):
        """
        Pause mapping.

        Args:
            entity_mapping_id: `str` or list of strings
                Mapping ids to be stopped.
            staging_name:
                Staging name to stop the mappings
            connector_name: `str`, `str`, default `None`
                Connector Name
            connector_id: `str`, `str`, default `None`
                Connector ID
            reverse_mapping: `bool` default `False`
                When using with consumer.False if you don't know what consumer is.

        Returns: dict
         Dictionary with the response of all mappings played.

        """

        responses = self._play_pause_mapping(
            kind='pause', entity_mapping_id=entity_mapping_id, staging_name=staging_name,
            connector_name=connector_name, connector_id=connector_id,
            reverse_mapping=reverse_mapping,
        )

        return responses

    def _play_pause_etl(self, kind, staging_name=None,
                        connector_name=None, connector_id=None, ):

        connector_id = connector_id if connector_id else self.get_by_name(connector_name)[
            'mdmId']
        resp = self.carol.call_api(
            path=f'v1/etl/staging/{connector_id}/{staging_name}/{kind}', method='POST')

        return resp

    def play_etl(
            self, staging_name=None,
            connector_name=None, connector_id=None,
    ):
        """
        Start ETL processes.

        Args:
            staging_name:
                Staging name for starting the ETLs
            connector_name: `str`, `str`, default `None`
                Connector Name
            connector_id: `str`, `str`, default `None`
                Connector ID

        Returns: dict
         Dictionary with the response of all mappings played.

        """

        responses = self._play_pause_etl(
            kind='play', staging_name=staging_name,
            connector_name=connector_name, connector_id=connector_id,
        )

        return responses

    def pause_etl(
            self, staging_name=None,
            connector_name=None, connector_id=None,
    ):
        """
        Pause ETL.

        Args:
            staging_name:
                Staging name to stop the ETLs
            connector_name: `str`, `str`, default `None`
                Connector Name
            connector_id: `str`, `str`, default `None`
                Connector ID

        Returns: dict
         Dictionary with the response Carol's response

        """

        responses = self._play_pause_etl(
            kind='pause', staging_name=staging_name,
            connector_name=connector_name, connector_id=connector_id,

        )

        return responses

    def get_etl_information(self, connector_id=None, connector_name=None):
        """
        Get ETL Configurations for a connector

        Args:

            connector_name: `str`, `str`, default `None`
                Connector Name
            connector_id: `str`, `str`, default `None`
                Connector ID

        Returns: list of dict
            Etl information

        """

        connector_id = connector_id if connector_id else self.get_by_name(connector_name)[
            'mdmId']

        return self.carol.call_api(path=f"v1/etl/connector/{connector_id}")

    def get_all_stagings(self, connector_name=None, connector_id=None):
        """
        Get all stagings from a connector.

        Args:

            connector_name: `str`, `str`, default `None`
                Connector Name
            connector_id: `str`, `str`, default `None`
                Connector ID

        Returns: list
            List of staging tables.

        """
        if connector_name is not None:
            connector_id = self.get_by_name(connector_name)['mdmId']
        elif connector_id is None:
            raise ValueError(
                'Either `connector_id` or `connnector_name` should be set.')

        c = self.carol.call_api(path=f"v1/staging/connectors/{connector_id}/tables")
        return sorted(c)



"""

This submodule has all the classes to query data from RT layer in Carol.

"""

import json
import itertools
from datetime import datetime
from .connectors import Connectors
from .named_query import NamedQuery
from .filter import Filter, MAXIMUM, MINIMUM, TYPE_FILTER, TERM_FILTER
from .filter import RANGE_FILTER as RF
from .utils.miscellaneous import ranges
import copy


def delete_golden(carol, dm_name, now=None):
    """
    Delete Golden records.

    It Will delete all golden records of a given data model based on lastUpdate.

    Args:

        carol: `pycarol.carol.Carol`
            Carol instance
        dm_name: `str`
            Data model name
        now: `str`
            Delete records where last update is less the `now`. Any date time ISO format is accepted.

    Usage:

    .. code:: python

        from pycarol.query import delete_golden
        from pycarol.auth.PwdAuth import PwdAuth
        from pycarol.carol import Carol
        login = Carol()
        delete_golden(login, dm_name=my_dm)

        #To delete based on a date.
        delete_golden(login, dm_name=my_dm, now='2018-11-16')

    Attention:

        This API will delete all data in the DataModel, and if there is a DataModel View related to this DataModel
        one needs to reprocess it.

    """

    if now is None:
        now = datetime.utcnow().isoformat(timespec='seconds')

    json_query = Filter.Builder() \
        .should(TYPE_FILTER(value=dm_name + "Golden")) \
        .should(TYPE_FILTER(value=dm_name + "Master")) \
        .must(RF("mdmLastUpdated", [None, now])) \
        .build().to_json()

    try:
        Query(carol).delete(json_query)
    except:
        # if it it too many records, one would have a timeout but the records will be deleted anyway
        pass

    json_query = Filter.Builder() \
        .type(dm_name + "Rejected") \
        .must(RF("mdmLastUpdated", [None, now])) \
        .build().to_json()

    try:
        Query(carol, index_type='STAGING').delete(json_query)
    except:
        # if it it too many records, one would have a timeout but the records will be deleted anyway
        pass


class Query:
    """
    Class to query data from Carol.

    This class can be used to query data from data models and stagings tables, query using named queries and delete
    records.

    Args:

        carol: carol: Carol object
            Carol object.
        max_hits:  `int`, default float('inf')
            Number of records that will be downloaded.
        offset: `int`, default 0
            Offset for pagination. Only used when `scrollable=False`
        page_size: `int`, default 100
            Number of records downloaded in each pagination. The maximum value is 1000
        sort_order: `str`, default 'ASC'
            Sort ascending ('ASC') vs. descending ('DESC').
        sort_by: `str`,  default `None`
            Name to sort by.
        scrollable: `bool`, default True
            Use scroll for pagination. This should be the main way of doing, unless you are querying few data.
        index_type: `str`, default 'MASTER'
            Query data from 'MASTER', 'STAGING'
        only_hits: `bool`, default 'True'
            Return only results in the response path $hits.mdmGoldenFieldAndValues
        fields: `list`, default `None`
            Fields to return in response. e.g., ["mdmGoldenFieldAndValues.mdmtaxid", "mdmGoldenFieldAndValues.date"]
        get_aggs: `bool`, default `False`
            To be used if the query/named query has aggravations
        save_results: `bool`, default `False`
            If save the result of the query in the file specified in `filename`
        filename: `str`, default `query_result.json`
            File path to save the response.
        print_status: `bool`, default `True`
            Print the number of records in each interaction.
        safe_check: `bool`, default `False`
            To be used if there are repeated records (same mdmId)
        get_errors: `bool`, default `False`
            To get the errors in the goldenRecords, if any.
        flush_result: `bool`, default `False`
            To be used with save_results, it will not copy the result to memory, only to the file.
        use_stream: `bool`, default `False`
            Use the stram of data.
        get_times: `bool`, default `False`
            It will create a list of times that each pagination took.
        kwargs: `dict`
            Extra parameters to be passed to Carol.call_api

    """
    def __init__(self, carol, max_hits=float('inf'), offset=0, page_size=100, sort_order='ASC', sort_by=None,
                 scrollable=True, index_type='MASTER', only_hits=True, fields=None, get_aggs=False,
                 save_results=False, filename='query_result.json', print_status=True, safe_check=False,
                 get_errors=False, flush_result=False, use_stream=False, get_times=False, **kwargs):

        self.carol = carol
        self.max_hits = max_hits
        self.offset = offset
        self.page_size = page_size
        self.sort_order = sort_order
        self.sort_by = sort_by
        self.scrollable = scrollable
        self.index_type = index_type
        self.only_hits = only_hits
        self.fields = fields
        self.get_aggs = get_aggs
        self.use_stream = use_stream

        self.save_results = save_results
        self.filename = filename
        self.print_status = print_status
        self.safe_check = safe_check
        self.get_errors = get_errors
        self.flush_result = flush_result
        self.get_times = get_times
        self.named_query = None
        self.callback = None

        # Crated to send to the Rest API
        self.query_params = None
        self.drop_list = None
        self.json_query = None
        self.total_hits = None
        self.query_times = []
        self.kwargs = kwargs

        self.results = []

        if self.max_hits == float('inf'):
            self.get_all = True
        else:
            self.get_all = False

    def _build_query_params(self):
        self.query_params = {"offset": self.offset, "pageSize": self.page_size, "sortOrder": self.sort_order,
                             "indexType": self.index_type}

        if self.sort_by is not None:
            self.query_params["sortBy"] = self.sort_by
        if self.scrollable:
            self.query_params["scrollable"] = self.scrollable
        if self.fields:
            self.query_params["fields"] = self.fields

    def _build_return_fields(self):
        if isinstance(self.fields, str):
            self.fields = [self.fields]

        if self.fields is not None:
            self.fields = ','.join(self.fields)


    def page(self, offset=0):
        """
        Get only one page of the result using offset.

        Args:
            offset: `int`, default 0
                Offset to get. To properly paginate manually, offset should be `offset + page_size`.

        Returns:
            Query json response
        """

        self.offset = offset
        self.scrollable = False

        self.results = []
        if self.json_query is None:
            raise ValueError("You must call query() or named() before calling page()")

        self._build_return_fields()
        self._build_query_params()

        if self.named_query is None:
            url_filter = "v2/queries/filter"
        else:
            url_filter = "v2/queries/named/{}".format(self.named_query)

        result = self.carol.call_api(url_filter, data=self.json_query, params=self.query_params, timeout=240,
                                     method_whitelist=frozenset(['POST']), **self.kwargs)

        if self.only_hits:

            result = result['hits']
            return [elem.get('mdmGoldenFieldAndValues', elem)
                    for elem in result if  elem.get('mdmGoldenFieldAndValues', None)]

        else:
            return result

    def go(self, callback=None):
        """

        Args:

            callback: `callable` object
                This function will receive the current batch of records from the filter made.
        Returns: `None`

        """

        self.results = []
        if self.json_query is None:
            raise ValueError("You must call all() or query() or named() before calling go()")

        self._build_return_fields()
        self._build_query_params()

        self._scrollable_query_handler(callback)

        return self


    def _scrollable_query_handler(self, callback=None):
        if not self.offset == 0:
            raise ValueError('It is not possible to use offset when using scroll for pagination')

        set_param = True
        count = self.offset
        if self.save_results:
            file = open(self.filename, 'w', encoding='utf8')

        if self.named_query is None:
            url_filter = "v2/queries/filter"
        else:
            url_filter = "v2/queries/named/{}".format(self.named_query)

        to_get = float("inf")
        downloaded = 0
        while count < to_get:

            result = self.carol.call_api(url_filter, data=self.json_query, params=self.query_params, timeout=240,
                                         method_whitelist=frozenset(['POST']), **self.kwargs)

            if set_param:
                self.total_hits = result["totalHits"]
                if self.get_all:
                    to_get = result["totalHits"]
                elif self.max_hits <= result["totalHits"]:
                    to_get = self.max_hits
                else:
                    to_get = result["totalHits"]

                set_param = False
                if self.safe_check:
                    self.mdmId_list = []
                if self.get_errors:
                    self.query_errors = {}

            count += result['count']
            downloaded += result['count']
            scroll_id = result.get('scrollId', None)
            if (scroll_id is not None) or (self.get_aggs):
                url_filter = "v2/queries/filter/{}".format(scroll_id)
            elif (result['count'] == 0):
                if count < self.total_hits:
                    print(f'Total records downloaded: {count}/{self.total_hits}')
                    print(f'Something is wrong, no scrollId to continue \n')
                    break
            else:
                raise Exception('No Scroll Id to use. Something is wrong')

            if self.get_times:
                self.query_times.append(result.pop('took'))

            if self.only_hits:

                result = result['hits']
                if self.safe_check:
                    self.mdmId_list.extend([mdm_id['mdmId'] for mdm_id in result])
                    if len(self.mdmId_list) > len(set(self.mdmId_list)):
                        raise Exception('There are repeated records')

                if self.get_errors:
                    self.query_errors.update(
                        {elem.get('mdmId', elem): elem.get('mdmErrors', elem) for elem in result if elem['mdmErrors']})

                result = [elem.get('mdmGoldenFieldAndValues', elem) for elem in result if
                          elem.get('mdmGoldenFieldAndValues',
                                   None)]  # get mdmGoldenFieldAndValues if not empty and if it exists

                if not self.flush_result:
                    self.results.extend(result)
            else:
                result.pop('count')
                result.pop('totalHits')
                # result.pop('scrollId')
                if not self.flush_result:
                    self.results.append(result)

                if self.get_aggs:
                    if self.save_results:
                        file.write(json.dumps(result, ensure_ascii=False))
                        file.write('\n')
                        file.flush()
                    break

            if callback:
                if callable(callback):
                    callback(result)
                else:
                    raise Exception(
                        f'"{callback}" is a {type(callback)} and is not callable. This variable must be a function.')

            if self.print_status:
                print('{}/{}'.format(downloaded, to_get), end='\r')
            if self.save_results:
                file.write(json.dumps(result, ensure_ascii=False))
                file.write('\n')
                file.flush()
        if self.save_results:
            file.close()

    def check_total_hits(self, json_query, index_type="MASTER"):
        """
        Check the total hits for a given query
        :param json_query: Json object with the query to use
        :param index_type: Index type to query.
        :return: number of records for this query
        """

        querystring = {"indexType": index_type}
        self.json_query = json_query
        url_filter = "v2/queries/count"
        result = self.carol.call_api(url_filter, data=self.json_query, method='POST', params=querystring,
                                     method_whitelist=frozenset(['HEAD', 'TRACE', 'GET',
                                                                 'PUT', 'OPTIONS', 'DELETE', 'POST']))
        self.total_hits = result
        return result

    def named(self, named_query, params=None, json_query=None):
        if json_query is not None and params is not None:
            raise Exception("You can only specify either params or json_query, but not both!")

        self.results = []
        self.named_query = named_query

        if json_query is not None:
            self.json_query = json_query
        else:
            if params is None:
                params = {}
            self.json_query = params

        return self

    def named_query_params(self, named_query):
        named = NamedQuery(self.carol)
        named.by_name(named_query=named_query)
        return named.param_dict

    def query(self, json_query):
        self.json_query = json_query
        return self

    def all(self, dm_name):
        if not dm_name.endswith('Golden'):
            dm_name = dm_name + 'Golden'
        self.json_query = Filter.Builder().type(dm_name).build().to_json()
        self.index_type = 'MASTER'
        return self

    def delete(self, json_query):
        # TODO: we should check the number of records to be deleted. If too many,
        # it can be a problem.
        self.json_query = json_query
        self.querystring = {"indexType": self.index_type}
        url_filter = "v2/queries/filter"
        result = self.carol.call_api(url_filter, data=self.json_query,
                                     params=self.querystring, method='DELETE',
                                     method_whitelist=frozenset(['HEAD', 'TRACE', 'GET',
                                                                 'PUT', 'OPTIONS']))

        print('Deleted: ', result)


class ParQuery:
    def __init__(self, carol, backend='dask', return_df=True, verbose=50, n_jobs=4):
        """

        :param carol:
        :param backend:
        :param return_df:
        :param verbose:
        :param n_jobs:
        """
        import pandas as pd
        self._stag_mdm_key_range = None
        self._multiplier = None
        self.carol = carol
        self.return_df = return_df
        if return_df:
            import pandas as pd
        self.backend = backend
        self.verbose = verbose
        self.n_jobs = n_jobs


        assert self.backend == 'dask' or self.backend == 'joblib'

    def _get_min_max(self, datamodel_name, mdm_key, index_type, custom_filter=None, multiplier=None ):

        if custom_filter is not None:
            j = custom_filter
        else:
            j = Filter.Builder() \
                .type(datamodel_name) \
                .aggregation_list([MINIMUM(name='MINIMUM', params=mdm_key), MAXIMUM(name='MAXIMUM', params=mdm_key)]) \
                .build().to_json()

        query = Query(self.carol, index_type=index_type, only_hits=False, get_aggs=True, save_results=False,
                      print_status=True, page_size=1, ).query(j).go()

        if query.results[0].get('aggs') is None:
            return None, None, None
        sample = query.results[0].get('hits')[0]
        min_v = query.results[0]['aggs']['MINIMUM']['value']
        max_v = query.results[0]['aggs']['MAXIMUM']['value']
        print(f"Total Hits to download: {query.total_hits}")

        if multiplier is not None:
            min_v = int(min_v*multiplier) - 10
            max_v = int(max_v*multiplier) + 10
        return min_v, max_v, sample

    def _get_staging_from_golden_rejected(self,datamodel_name, connector_id, staging_name, fields):

        index_type = 'STAGING'
        self.datamodel_name = f"{datamodel_name}Rejected"
        self.fields = 'mdmStagingRecord'
        self.filter_stag = f"{connector_id}_{staging_name}"
        only_hits = False
        mdm_key = 'mdmStagingRecord.mdmCounterForEntity'
        j = Filter.Builder() \
            .type(self.datamodel_name) \
            .must(TERM_FILTER(key='mdmStagingEntityName.raw',
                              value=self.filter_stag)) \
            .aggregation_list([MINIMUM(name='MINIMUM', params=mdm_key),
                               MAXIMUM(name='MAXIMUM', params=mdm_key)]) \
            .build().to_json()

        min_v, max_v, sample = self._get_min_max(datamodel_name=self.datamodel_name, mdm_key=mdm_key,
                                                 index_type=index_type, custom_filter=j)
        if (min_v is None) and (max_v is None):
            return []
        chunks = ranges(min_v, max_v, self.slices)

        # rejected

        print(f"Number of chunks for rejected: {len(chunks)}")
        self.fields_to_get = [self.fields + '.' + i for i in sample.get(self.fields).keys() for j in fields if
                              j + '_' in i]

        self.custom_filter = Filter.Builder() \
            .type(self.datamodel_name) \
            .must(TERM_FILTER(key='mdmStagingEntityName.raw',
                              value=self.filter_stag))

        if self.backend == 'dask':
            list_to_compute = _dask_backend(carol=self.carol, chunks=chunks, datamodel_name=self.datamodel_name,
                                            page_size=self.page_size, index_type=index_type, fields=self.fields,
                                            only_hits=only_hits, mdm_key=mdm_key, return_df=self.return_df,
                                            fields_to_get=self.fields_to_get, custom_filter=self.custom_filter)
        elif self.backend == 'joblib':
            list_to_compute = _joblib_backend(carol=self.carol, chunks=chunks, datamodel_name=self.datamodel_name,
                                              page_size=self.page_size, index_type=index_type, fields=self.fields,
                                              only_hits=only_hits, mdm_key=mdm_key, return_df=self.return_df,
                                              fields_to_get=self.fields_to_get, custom_filter=self.custom_filter,
                                              n_jobs=self.n_jobs, verbose=self.verbose)
        else:
            raise KeyError

        return list_to_compute



    def _get_staging_from_golden(self, datamodel_name=None, staging_name=None, fields=None,
                                 connector_id=None, connector_name=None):
        cc = Connectors(self.carol)
        st = cc.get_dm_mappings(all_connectors=True)

        if datamodel_name is None:
            assert staging_name is not None, "staging_name should be set."
            if (connector_id is None) and (connector_name is not None):
                connector_id = Connectors(self.carol).get_by_name(connector_name)['mdmId']
                entity = [i['mdmMasterEntityName'] for i in st
                          if (i.get('mdmConnectorId') == connector_id) and
                          (i.get('mdmStagingType') == staging_name)]
                assert len(entity) == 1, (f'No data model mapped for {staging_name}')
                datamodel_name = entity[0]

            elif connector_id is None:
                entity = [i for i in st if (i.get('mdmStagingType') == staging_name)]

                if len(entity) > 1:
                    raise KeyError(f'There are more than one connector for staging {staging_name}')
                elif len(entity) < 1:
                    raise KeyError(f'No data model mapped for {staging_name}')
                entity = entity[0]
                connector_id = entity['mdmConnectorId']
                datamodel_name = entity['mdmMasterEntityName']
            elif connector_id:
                entity = [i['mdmMasterEntityName'] for i in st
                          if (i.get('mdmConnectorId') == connector_id) and
                          (i.get('mdmStagingType') == staging_name)]
                assert len(entity) == 1, (f'No data model mapped for {staging_name}')
                datamodel_name = entity[0]

        elif staging_name is None:
            entity = [i for i in st
                      if (i.get('mdmMasterEntityName') == datamodel_name)]
            if len(entity) > 1:
                raise KeyError(
                    f'There are more than one staging mapped for the data model {datamodel_name}\n Use "staging_name" for disambiguation ')
            elif len(entity) < 1:
                raise KeyError(f'No mapping fo data model {datamodel_name}')
            entity = entity[0]
            connector_id = entity['mdmConnectorId']
            staging_name = entity['mdmStagingType']
        else:
            assert staging_name is not None, "staging_name should be set."
            entity = [i for i in st if (i.get('mdmStagingType') == staging_name)]
            if len(entity) > 1:
                raise KeyError(
                    f'There are more than one connector with this staging name\n Use "connector_id" for disambiguation ')
            elif len(entity) < 1:
                raise KeyError(f'No mapping for data model {datamodel_name}')

            entity = entity[0]
            connector_id = entity['mdmConnectorId']
            staging_name = staging_name

        index_type = 'MASTER'
        self.datamodel_name = f"{datamodel_name}Master"
        self.fields = 'mdmStagingRecord'
        self.filter_stag = f"{connector_id}_{staging_name}"
        only_hits = False
        mdm_key = 'mdmCounterForEntity'
        j = Filter.Builder() \
            .type(self.datamodel_name) \
            .must(TERM_FILTER(key='mdmStagingEntityName.raw',
                              value=self.filter_stag)) \
            .aggregation_list([MINIMUM(name='MINIMUM', params=mdm_key),
                               MAXIMUM(name='MAXIMUM', params=mdm_key)]) \
            .build().to_json()

        min_v, max_v, sample = self._get_min_max(datamodel_name=self.datamodel_name, mdm_key=mdm_key,
                                                 index_type=index_type, custom_filter=j)
        if (min_v is None) and (max_v is None):
            return []
        chunks = ranges(min_v, max_v, self.slices)

        print(f"Number of chunks: {len(chunks)}")
        print("Getting staging from Golden, after will get from rejected too. ")
        self.fields_to_get = [self.fields + '.' + i for i in sample.get(self.fields).keys() for j in fields if
                              j + '_' in i]

        self.custom_filter = Filter.Builder() \
            .type(self.datamodel_name) \
            .must(TERM_FILTER(key='mdmStagingEntityName.raw',
                              value=self.filter_stag))

        if self.backend == 'dask':
            list_to_compute = _dask_backend(carol=self.carol, chunks=chunks, datamodel_name=self.datamodel_name,
                                            page_size=self.page_size, index_type=index_type, fields=self.fields,
                                            only_hits=only_hits, mdm_key=mdm_key, return_df=self.return_df,
                                            fields_to_get=self.fields_to_get, custom_filter=self.custom_filter)
        elif self.backend == 'joblib':
            list_to_compute = _joblib_backend(carol=self.carol, chunks=chunks, datamodel_name=self.datamodel_name,
                                              page_size=self.page_size, index_type=index_type, fields=self.fields,
                                              only_hits=only_hits, mdm_key=mdm_key, return_df=self.return_df,
                                              fields_to_get=self.fields_to_get, custom_filter=self.custom_filter,
                                              n_jobs=self.n_jobs, verbose=self.verbose)
        else:
            raise KeyError

        list_to_compute_rejected = self._get_staging_from_golden_rejected(datamodel_name=datamodel_name,
                                                                          connector_id=connector_id,
                                                                          staging_name=staging_name,
                                                                          fields=fields)
        print("Getting staging from rejected")
        list_to_compute.extend(list_to_compute_rejected)
        if self.return_df:
            return pd.concat(list_to_compute, ignore_index=True, sort=True)
        list_to_compute = list(itertools.chain(*list_to_compute))

        return list_to_compute


    def _get_golden(self, datamodel_name=None, fields=None):

        index_type = 'MASTER'
        self.datamodel_name = f"{datamodel_name}Golden"
        self.fields = 'mdmGoldenFieldAndValues'
        self.fields_to_get = [self.fields + '.' + i if self.fields not in i else i for i in fields]
        only_hits = True
        mdm_key = 'mdmCounterForEntity'

        min_v, max_v, sample = self._get_min_max(datamodel_name=self.datamodel_name, mdm_key=mdm_key,
                                                 index_type=index_type)
        if (min_v is None) and (max_v is None):
            return []
        chunks = ranges(min_v, max_v, self.slices)
        print(f"Number of chunks: {len(chunks)}")

        if self.backend == 'dask':
            list_to_compute = _dask_backend(carol=self.carol, chunks=chunks, datamodel_name=self.datamodel_name,
                                            page_size=self.page_size, index_type=index_type, fields=self.fields,
                                            only_hits=only_hits, mdm_key=mdm_key, return_df=self.return_df,
                                            fields_to_get=self.fields_to_get, custom_filter=self.custom_filter)
        elif self.backend == 'joblib':
            list_to_compute = _joblib_backend(carol=self.carol, chunks=chunks, datamodel_name=self.datamodel_name,
                                              page_size=self.page_size, index_type=index_type, fields=self.fields,
                                              only_hits=only_hits, mdm_key=mdm_key, return_df=self.return_df,
                                              fields_to_get=self.fields_to_get, custom_filter=self.custom_filter,
                                              n_jobs=self.n_jobs, verbose=self.verbose)

        if self.return_df:
            return pd.concat(list_to_compute, ignore_index=True, sort=True)
        list_to_compute = list(itertools.chain(*list_to_compute))
        return list_to_compute


    def _get_staging(self, connector_id=None, connector_name=None, staging_name=None, fields=None):

        if not connector_id:
            connector_id = Connectors(self.carol).get_by_name(connector_name)['mdmId']

        index_type = 'STAGING'
        self.datamodel_name = f"{connector_id}_{staging_name}"
        self.fields_to_get = fields
        self.fields = None
        only_hits = False
        mdm_key = 'mdmCounterForEntity'
        if self._stag_mdm_key_range is not None:
            mdm_key = self._stag_mdm_key_range



        min_v, max_v, sample = self._get_min_max(datamodel_name=self.datamodel_name, mdm_key=mdm_key,
                                                 index_type=index_type, multiplier=self._multiplier )
        if (min_v is None) and (max_v is None):
            return []
        chunks = ranges(min_v, max_v, self.slices)
        print(f"Number of chunks: {len(chunks)}")

        if self.backend == 'dask':
            list_to_compute = _dask_backend(carol=self.carol, chunks=chunks, datamodel_name=self.datamodel_name,
                                            page_size=self.page_size, index_type=index_type, fields=self.fields,
                                            only_hits=only_hits, mdm_key=mdm_key, return_df=self.return_df,
                                            fields_to_get=self.fields_to_get, custom_filter=self.custom_filter)
        elif self.backend == 'joblib':
            list_to_compute = _joblib_backend(carol=self.carol, chunks=chunks, datamodel_name=self.datamodel_name,
                                              page_size=self.page_size, index_type=index_type, fields=self.fields,
                                              only_hits=only_hits, mdm_key=mdm_key, return_df=self.return_df,
                                              fields_to_get=self.fields_to_get, custom_filter=self.custom_filter,
                                              n_jobs=self.n_jobs, verbose=self.verbose)
        else:
            raise KeyError

        if self.return_df:
            return pd.concat(list_to_compute, ignore_index=True, sort=True)
        list_to_compute = list(itertools.chain(*list_to_compute))
        return list_to_compute



    def go(self, datamodel_name=None, slices=1000, page_size=1000, staging_name=None, connector_id=None,
           connector_name=None,
           get_staging_from_golden=False, fields=None):
        assert slices < 9999, '10k is the largest slice possible'
        self.slices = slices
        self.page_size = page_size
        if fields is None:
            fields = []
        self.custom_filter = None

        if get_staging_from_golden:
            list_to_compute = self._get_staging_from_golden(datamodel_name=datamodel_name, staging_name=staging_name,
                                                            connector_id=connector_id, connector_name=connector_name,
                                                            fields=fields)

            return list_to_compute

        if datamodel_name is None:
            assert connector_id or connector_name
            assert staging_name

            return self._get_staging(connector_id=connector_id, connector_name=connector_name, staging_name=staging_name,
                              fields=fields)
        else:

            return self._get_golden(datamodel_name=datamodel_name, fields=fields)


def _dask_backend(carol, chunks, datamodel_name, page_size, index_type, fields,
                  only_hits, mdm_key, return_df, fields_to_get, custom_filter):
    list_to_compute = []
    for RANGE_FILTER in chunks:
        y = dask.delayed(_par_query)(
            datamodel_name=datamodel_name,
            RANGE_FILTER=RANGE_FILTER,
            page_size=page_size,
            login=carol,
            index_type=index_type,
            fields=fields,
            only_hits=only_hits,
            mdm_key=mdm_key,
            return_df=return_df,
            fields_to_get=fields_to_get,
            custom_filter=custom_filter,
        )
        list_to_compute.append(y)

    return dask.compute(*list_to_compute)


def _joblib_backend(carol, chunks, datamodel_name, page_size, index_type, fields,
                    only_hits, mdm_key, return_df, fields_to_get, custom_filter, n_jobs, verbose, ):
    from joblib import Parallel, delayed
    list_to_compute = Parallel(n_jobs=n_jobs,
                               verbose=verbose)(delayed(_par_query)(
        datamodel_name=datamodel_name,
        RANGE_FILTER=RANGE_FILTER,
        page_size=page_size,
        login=carol,
        index_type=index_type,
        fields=fields,
        only_hits=only_hits,
        mdm_key=mdm_key,
        return_df=return_df,
        fields_to_get=fields_to_get,
        custom_filter=custom_filter,
    )
                                                for RANGE_FILTER in chunks)
    return list_to_compute


def _par_query(datamodel_name, RANGE_FILTER, page_size=1000, login=None, index_type='MASTER', fields=None, mdm_key=None,
               only_hits=True, return_df=True, fields_to_get=None, custom_filter=None):
    if custom_filter is not None:
        json_query = copy.deepcopy(custom_filter)
        json_query = json_query.must(RF(key=mdm_key, value=RANGE_FILTER)).build().to_json()

    else:
        json_query = Filter.Builder() \
            .type(datamodel_name) \
            .must(RF(key=mdm_key, value=RANGE_FILTER)) \
            .build().to_json()

    query = Query(login, page_size=page_size, save_results=False, print_status=False, index_type=index_type,
                  only_hits=only_hits,
                  fields=fields_to_get).query(json_query).go()
    query = query.results

    if not only_hits:
        query = [i['hits'] for i in query]
        query = list(itertools.chain(*query))
        if fields:
            query = [elem.get(fields, elem) for elem in query if
                     elem.get(fields, None)]

    if return_df:
        return pd.DataFrame(query)

    return query



import os
from contextlib import contextmanager


@contextmanager
def dotenv_context():
    """``with`` context to temporarily modify the environment variables"""
    import os, dotenv
    _environ = os.environ.copy()
    try:
        dotenv.load_dotenv()
        yield
    finally:
        os.environ.clear()
        os.environ.update(_environ)


def assert_env_is_defined(needed_keys=frozenset(['CAROLTENANT', 'CAROLAPPVERSION', 'CAROLAPPNAME',
                                                 'CAROLAPPOAUTH', 'CAROLCONNECTORID'])
                          ):
    """Raises KeyError if obligatory env variables were not found"""
    key_exists = {k: False for k in needed_keys}
    for k in needed_keys:
        key_exists[k] = k in os.environ.keys()
    if not all(key_exists.values()):
        raise KeyError(f"some env variables were not defined: {key_exists}")



from urllib3.util.retry import Retry
import requests
from requests.adapters import HTTPAdapter
import json
import os, copy
import os.path
from .auth.ApiKeyAuth import ApiKeyAuth
from .auth.PwdAuth import PwdAuth
from .tenant import Tenant
from . import __CONNECTOR_PYCAROL__
from . import __version__
from . organization import Organization
from .exceptions import CarolApiResponseException, InvalidToken

class Carol:
    """
    Handles all Carol`s API calls It will handle all API calls,
    for a given authentication method. :param domain: `str`.

    Args:

        domain: `str`. default `None`.
            Tenant name. e.x., domain.carol.ai
        app_name: `str`. default `None`.
            Carol app name.
        auth: `PwdAuth` or `ApiKeyAuth`.
            object Auth Carol object to handle authentication
        connector_id: `str` , default `__CONNECTOR_PYCAROL__`.
            Connector Id
        port: `int` , default 443.
            Port to be used (when running locally it could change)
        verbose: `bool` , default `False`.
            If True will print the header, method and URL of each API call.
        organization: `str` , default `None`.
            Organization domain.
        environment: `str`, default `carol.ai`,
            Which Carol's environment to use. There are three possible values today.

                1. 'carol.ai' for the production environment
                2. 'karol.ai' for the explore environment
                3. 'qarol.ai' for the QA environment

        host: `str` default `None`
            This will overwrite the host used. Today the host is:

                1. if organization is None, host={domain}.{environment}
                2. else host={organization}.{environment}

            See Carol._set_host.
        
         user:  `str` default `None`
            User
         password: `str` default `None`
            User passowrd
         api_key:  `str` default `None`
            Carol's Api Key

    OBS:
        In case all parameters are `None`, pycarol will try yo find their values in the environment variables.
        The values are:

             1. `CAROLTENANT` for domain
             2. `CAROLAPPNAME` for app_name
             3. `CAROLAPPOAUTH` for auth
             4. `CAROLORGANIZATION` for organization
             5. `CAROLCONNECTORID` for connector_id
             6. `CAROL_DOMAIN` for environment
             7. `CAROLUSER` for carol user email
             8. `CAROLPWD` for user password.


    """

    def __init__(self, domain=None, app_name=None, auth=None, connector_id=None, port=443, verbose=False,
                 organization=None, environment=None, host=None, user=None, password=None, api_key=None):

        self.connector_id = connector_id

        if auth is None:
            auth = self.__get_auth(connector_id=connector_id, username=user, password=password, app_oauth=api_key)

        if domain is None:
            domain = os.getenv('CAROLTENANT')
            if domain is None:
                raise ValueError("`domain` must be set.")

        if app_name is None:
            app_name = os.getenv('CAROLAPPNAME', ' ')

        if self.connector_id is None:
            if auth.connector_id is None:
                self.connector_id = __CONNECTOR_PYCAROL__
            else:
                self.connector_id = auth.connector_id

        if domain is None or app_name is None or auth is None:
            raise ValueError("domain, app_name and auth must be specified as parameters, either " +
                             "in the environment variables CAROLTENANT, CAROLAPPNAME, CAROLAPPOAUTH" +
                             " or CAROLUSER+CAROLPWD and  CAROLCONNECTORID")

        # TODO Fixed to be compatible with the old `ENV_DOMAIN`. We could add a deprecated warning.
        self.environment = environment if environment is not None else os.getenv('CAROL_DOMAIN',
                                                                                 os.getenv('ENV_DOMAIN', 'carol.ai'))
        self.organization = organization if organization is not None else os.getenv('CAROLORGANIZATION',
                                                                                    os.getenv('API_SUBDOMAIN'))
        self.domain = domain
        self.app_name = app_name
        self.port = port
        self.verbose = verbose
        self._host_string = host
        self.host = self._set_host(domain=self.domain, organization=self.organization,
                                   environment=self.environment, host=host)
        self._tenant = None
        self.auth = auth
        self.auth.set_connector_id(self.connector_id)
        self.auth.login(self)
        self.response = None

        self.org = None

    def __get_auth(self, connector_id=None, username=None, password=None, app_oauth=None):

        if username is None:
            username = os.getenv('CAROLUSER')

        if password is None:
            password = os.getenv('CAROLPWD')

        if username and password:
            return PwdAuth(user=username, password=password)
            
        if app_oauth is None:
            app_oauth = os.getenv('CAROLAPPOAUTH')
        
        if connector_id is None:
            connector_id = os.getenv('CAROLCONNECTORID', __CONNECTOR_PYCAROL__)
            self.connector_id = connector_id
        
        if app_oauth and connector_id:
            return ApiKeyAuth(app_oauth)

        raise ValueError("either `auth` or `username/password` or `api_key` or pycarol env variables must be set.")


    @property
    def tenant(self):
        if self._tenant is None:
            self._tenant = Tenant(self).get_tenant_by_domain(self.domain)
        return self._tenant


    @staticmethod
    def _set_host(domain, organization, environment, host):
        """
        Set the host to be used.

        Args:

            domain: `str`
                Former tenant name.
                e.x., domain.carol.ai
            organization: `str`
                Organization domain.
            environment: `str`
                Which Carol's environment to use. There are three possible values today.

                1. 'carol.ai' for the production environment
                2. 'karol.ai' for the explore environment
                3. 'qarol.ai' for the QA environment

            host: `str`
                Host to be used. It overwrite the default one.

        Returns: `str`
            host
        """
        if host is not None:
            return host
        elif organization is not None:
            return f"{organization}.{environment}"
        else:
            return f"{domain}.{environment}"
        pass

    @staticmethod
    def _retry_session(retries=5, session=None, backoff_factor=0.5, status_forcelist=(500, 502, 503, 504, 524),
                       method_whitelist=frozenset(['HEAD', 'TRACE', 'GET', 'PUT', 'OPTIONS', 'DELETE'])):

        """
        Static method used to handle retries between calls.

        Args:

            retries: `int` , default `5`
                Number of retries for the API calls
            session: Session object dealt `None`
                It allows you to persist certain parameters across requests.
            backoff_factor: `float` , default `0.5`
                Backoff factor to apply between  attempts. It will sleep for:
                        {backoff factor} * (2 ^ ({retries} - 1)) seconds
            status_forcelist: `iterable` , default (500, 502, 503, 504, 524).
                A set of integer HTTP status codes that we should force a retry on.
                A retry is initiated if the request method is in method_whitelist and the response status code is in
                status_forcelist.
            method_whitelist: `iterable` , default frozenset(['HEAD', 'TRACE', 'GET', 'PUT', 'OPTIONS', 'DELETE']))
                Set of uppercased HTTP method verbs that we should retry on.

        Returns:
            :class:`requests.Section`
        """

        session = session or requests.Session()
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
            method_whitelist=method_whitelist,
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount('http://', adapter)
        session.mount('https://', adapter)
        return session

    def call_api(self, path, method=None, data=None, auth=True, params=None, content_type='application/json', retries=8,
                 session=None, backoff_factor=0.5, status_forcelist=(502, 503, 504, 524), downloadable=False,
                 method_whitelist=frozenset(['HEAD', 'TRACE', 'GET', 'PUT', 'OPTIONS', 'DELETE', 'POST']), errors='raise',
                 extra_headers=None, files=None,
                 **kwds):
        """
        This method handles all the API calls.

        Args:

            path: `str`.
                API URI path. e.x.  v2/staging/schema
            method: 'str', default `None`.
                Set of uppercased HTTP method verbs that we should call on.
            data: 'dict`, default `None`.
                Dictionary, list of tuples, bytes, or file-like object to send in
                the body of the request.
            auth: :class: `pycarol.ApiKeyAuth` or `pycarol.PwdAuth`
                Auth type to be used within the API's calls.
            params: (optional) Dictionary, list of tuples or bytes to send
                     in the query string for the :class:`requests.Request`.
            content_type: `str`, default 'application/json'
                Content type for the api call
            retries: `int` , default `5`
                Number of retries for the API calls
            session: :class `requests.Session` object dealt `None`
                It allows you to persist certain parameters across requests.
            backoff_factor: `float` , default `0.5`
                Backoff factor to apply between  attempts. It will sleep for:
                        {backoff factor} * (2 ^ ({retries} - 1)) seconds
            status_forcelist: `iterable` , default (500, 502, 503, 504, 524).
                A set of integer HTTP status codes that we should force a retry on.
                A retry is initiated if the request method is in method_whitelist and the response status code is in
                status_forcelist.
            downloadable: `bool` default `False`.
                If the request will return a file to download.
            method_whitelist: `iterable` , default frozenset(['HEAD', 'TRACE', 'GET', 'PUT', 'OPTIONS', 'DELETE']))
                Set of uppercased HTTP method verbs that we should retry on.
            errors: {‘ignore’, ‘raise’}, default ‘raise’
                If ‘raise’, then invalid request will raise an exception If ‘ignore’,
                then invalid request will return the request response
            extra_headers: `dict` default `None`
                extra headers to be sent.
            files: `dict` default `None`
                Used when uploading files to carol. This will be sent to :class: `requests.request`
            kwds: `dict` default `None`
                Extra parameters to be sent to :class: `requests.request`

        Rerturn:
            Dict with API response.

        """

        extra_headers = extra_headers or {}
        url = f'https://{self.host}:{self.port}/api/{path}'

        if method is None:
            if data is None:
                method = 'GET'
            else:
                method = 'POST'

        met_list = ['HEAD', 'TRACE', 'GET', 'PUT','POST', 'OPTIONS', 'PATCH',
                    'DELETE', 'CONNECT' ]
        assert method in met_list, f"API method must be {met_list}"

        headers = {'accept': 'application/json'}
        if auth:
            self.auth.authenticate_request(headers)

        data_json = None
        if method == 'GET':
            pass

        elif (method == 'POST') or (method == 'DELETE') or (method == 'PUT'):
            if content_type is not None:
                headers['content-type'] = content_type

            if content_type == 'application/json':
                data_json = data
                data = None

        headers.update(extra_headers)
        headers.update({'User-Agent': f'pyCarol/{__version__}'})

        __count = 0
        while True:
            if session is None:
                session = self._retry_session(retries=retries, session=session, backoff_factor=backoff_factor,
                                              status_forcelist=status_forcelist, method_whitelist=method_whitelist)

            response = session.request(method=method, url=url, data=data, json=data_json,
                                       headers=headers, params=params, files=files, **kwds)

            if self.verbose:
                if data_json is not None:
                    print("Calling {} {}. Payload: {}. Params: {}".format(method, url, data_json, params))
                else:
                    print("Calling {} {}. Payload: {}. Params: {}".format(method, url, data, params))
                print("        Headers: {}".format(headers))

            if response.ok or errors == 'ignore':
                if downloadable:  #Used when downloading carol app file.
                    return response

                response.encoding = 'utf-8'
                self.response = response
                if response.text == '':
                    return {}
                return json.loads(response.text)

            elif (response.reason == 'Unauthorized') and isinstance(self.auth, PwdAuth):
                if response.json().get('possibleResponsibleField') in ['password', 'userLogin']:
                    raise InvalidToken(response.text)
                self.auth.get_access_token()  #It will refresh token if Unauthorized
                __count += 1
                if __count < 5: #To avoid infinity loops
                    continue
                else:
                    raise Exception('Too many retries to refresh token.\n', response.text, response.status_code)
            elif (response.status_code == 404):
                raise CarolApiResponseException(response.text, response.status_code)

            raise Exception(response.text, response.status_code)

    def issue_api_key(self, connector_id=None):
        """
        Create an API key for a given connector.

        Args:

            connector_id: `str` default `None`.
                Connector ID to be used when creating the APIkey

        Returns: `dict`
            Dictionary with the API key.

        """

        if connector_id is None:
            connector_id = self.connector_id
        resp = self.call_api('v2/apiKey/issue', data={
            'connectorId': connector_id
        }, content_type='application/x-www-form-urlencoded')
        return resp

    def api_key_details(self, api_key, connector_id):

        """
        Display information about the API key.

        Args:

            api_key: `str`
                Carol's api key
            connector_id: `str`
                Connector Id which API key was created.

        Returns: `dict`
            Dictionary with API key information.

        """

        resp = self.call_api('v2/apiKey/details',
                             params = {"apiKey": api_key, "connectorId": connector_id})

        return resp

    def api_key_revoke(self, connector_id):
        """
        Revoke API key for ta given connector_id

        Args:

            connector_id: `str`
                Connector Id which API key was created.

        Returns: `dict`

            Dictionary with API request response.

        """

        resp = self.call_api('v2/apiKey/revoke', method='DELETE',
                             content_type='application/x-www-form-urlencoded',
                             params = {"connectorId": connector_id})

        return resp

    def switch_org_level(self):

        org = self._current_org()
        self.auth.switch_org_context(org['mdmId'])

    def switch_environment(self, env_name=None, env_id=None, app_name=None, org_name=None, org_id=None):
        """
        Switch org/environments. If the user has access to this environment, it will be "logged in" in this new
        org/environment.

        Args:

            env_name: `str` default `None`
                Environment (tenant) name to switch the context to.
            env_id: `str` default `None`
                Environment (tenant) id to switch the context to.
            app_name: `str` default `None`
                App name in the target environment to switch the context to.
                Only needed with using CDS.
            org_name: `str` default `None`
                The organization name to switch context to. If the same keep it `None`
            org_id: `str` default `None`
                The organization id to switch context to. If the same keep it `None`

        Returns:
            self

        .. code:: python

            from pycarol import Carol, Staging
            carol = Carol('B', 'teste', auth=PwdAuth('email@totvs.com.br', 'pass'), )
            carol.switch_environment('A')
            Staging(carol_tenant_A).fetch_parquet(...) # fetch parquet from tenant A
            # To switch back
            carol.switch_environment('B')
            #back to tenant B


        """

        if self.org is None:
            self.org = Organization(self).get_organization_info(self.organization)

        current = self.get_current()

        if (org_id is not None and org_id != current['org_id']) or (org_name is not None and org_name != current['org_name'] ):
            # Switch to org context.
            self.switch_org_level()
            if org_id is None:
                org_id = Organization(self).get_organization_info(org_name)['mdmId']
            # Switch org.
            self.auth.switch_org_context(org_id)

        if env_name:
            env_id = Tenant(self).get_tenant_by_domain(env_name)['mdmId']
        elif env_id is None:
            raise ValueError('Either `env_name` or `env_id` must be set.')

        self.auth.switch_context(env_id=env_id)

        self.domain = env_name
        self.app_name = app_name # TODO: Today we cannot use CDS without a valid app name.
        self._tenant = self._current_env()
        self.organization = self._current_org()['mdmName']

        self.host = self._set_host(domain=self.domain, organization=self.organization,
                                   environment=self.environment, host=self._host_string)

        return self

    def switch_context(self, env_name=None, env_id=None, app_name=None, org_name=None, org_id=None):
        """
        Context manager to temporary have access to a second environment

        Args:

            env_name: `str` default `None`
                Environment (tenant) name to switch the context to.
            env_id: `str` default `None`
                Environment (tenant) id to switch the context to.
            app_name: `str` default `None`
                App name in the target environment to switch the context to.
                Only needed with using CDS.

        Returns:
            None

        Usage:

        .. code:: python

            from pycarol import Carol, Staging
            carol = Carol('B', 'teste', auth=PwdAuth('email@totvs.com.br', 'pass'), )
            with carol.switch_context('A') as carol_tenant_A:
                Staging(carol_tenant_A).fetch_parquet(...) # fetch parquet from tenant A
            #back to tenant B


        """

        if self.org is None:
            self.org = Organization(self).get_organization_info(self.organization)

        if env_name:
            env_id = Tenant(self).get_tenant_by_domain(env_name)['mdmId']
        elif env_id is None:
            raise ValueError('Either `env_name` or `env_id` must be set.')

        class SwitchContext(object):

            def __init__(self, parent_context, env_name=None, env_id=None, org_id=None, org_name=None, app_name=None):
                self.parent_context = parent_context
                self.env_name = env_name
                self.env_id = env_id
                self.app_name = app_name
                self.org_id = org_id
                self.org_name = org_name

            def __enter__(self):
                self.parent_context.switch_environment(env_name=self.env_name, env_id=self.env_id,
                                                       app_name=self.app_name, org_name=self.org_name,
                                                       org_id=self.org_id)
                return self.parent_context

            def __exit__(self, exc_type, exc_val, exc_tb):
                del self.parent_context

        return SwitchContext(parent_context=copy.deepcopy(self), env_name=env_name, env_id=env_id, app_name=app_name,
                             org_name=org_name, org_id=org_id,)


    def _current_env(self):
        return self.call_api('v1/tenants/current')

    def _current_org(self):
        return self.call_api('v1/organizations/current')

    def get_current(self, level='all'):
        """
        Get current org/env information.

        Args:

            level: `str`:
                Possible Values:
                    "all": To get organization and environment information.
                    "org": To get organization information.
                    "env": To get environment information.

        Returns: `dict`
            Dictionary with keys org_Id, org_name, env_id, env_name

        """

        env = {}
        org = {}

        if level.lower() not in ['org', 'env', 'all']:
            raise ValueError(f"level should be 'all', 'org', 'env', {level} was passed.")

        if level.lower() == 'env' or level.lower() == 'all':
            env = self._current_env()

        if level.lower() == 'org' or level.lower() == 'all':
            org = self._current_org()

        return {
            "env_name": env.get('mdmName'),
            "env_id": env.get('mdmId'),
            "org_name": org.get('mdmName'),
            "org_id": org.get('mdmId'),
        }



class CarolApiResponseException(Exception):
    """Custom Exception to handle Exception on Carol API Calls

    Args:
        Exception (Exception): SuperClass
    """
    pass

class InvalidToken(Exception):

    """Custom Exception to handle Invalid token
    Args:
        Exception (Exception): SuperClass
    """
    pass




import json
from datetime import datetime

class IntType(object):
    json_type = "integer"


class DoubleType(object):
    json_type = "double"


class StringType(object):
    json_type = "string"


class NullType(object):
    json_type = "string"


class BooleanType(object):
    json_type = "boolean"


class ArrayType(object):
    json_type = "object"
    items = []


class ObjectType(object):
    json_type = "object"
    properties = {}


class DateType(object):
    json_type = "date"

class Type(object):
    @classmethod
    def get_schema_type_for(cls, t):
        import pandas as pd
        import numpy as np

        """docstring for get_schema_type_for"""

        SCHEMA_TYPES = {
            type(None): NullType,
            str: StringType,
            int: IntType,
            float: DoubleType,
            bool: BooleanType,
            list: ArrayType,
            dict: ObjectType,
            np.float64: DoubleType,
            np.float32: DoubleType,
            np.float: DoubleType,
            np.int64: IntType,
            np.int32: IntType,
            np.int: IntType,
            np.bool_: BooleanType,
            datetime: DateType,
            pd.Timestamp:DateType
        }

        schema_type = SCHEMA_TYPES.get(t)

        if not schema_type:
            raise JsonSchemaTypeNotFound("There is no schema type for  %s.\n Try:\n %s" % (
            str(t), ",\n".join(["\t%s" % str(k) for k in SCHEMA_TYPES.keys()])))
        return schema_type


class JsonSchemaTypeNotFound(Exception):
    pass


class carolSchemaGenerator(object):
    def __init__(self, base_object):
        self.base_object = base_object
        self.schema_dict = None

    @classmethod
    def from_json(cls, base_json):
        base_object = json.loads(base_json)
        obj = cls(base_object)
        return obj

    def to_dict(self, mdmStagingType='stagingName', mdmFlexible=False, crosswalkname=None,
                crosswalkList=None, export_data=False):

        crosswalkname = mdmStagingType  #TODO. We can have more than 1 crosswalk, and one HAS to be named as the mdmStagingType
                                        # this is to enforce we use the same name

        assert isinstance(crosswalkList, list)


        fields = set(self.base_object.keys())
        for key_field in crosswalkList:
            if key_field not in fields:
                raise Exception('Your key field %s is not in your fields!' % (key_field))
            if type(key_field) != str:
                raise Exception('Field %s type is not string, it must be!' % key_field)

        if crosswalkname is None:
            crosswalkname = mdmStagingType

        mdmCrosswalkTemplate = {"mdmCrosswalkTemplate": {"mdmCrossreference": {crosswalkname: crosswalkList}}}

        schema_dict = {}

        base_object = self.base_object
        schema_dict["mdmStagingType"] = mdmStagingType
        schema_dict.update(mdmCrosswalkTemplate)
        schema_dict.update({"mdmStagingMapping": {"properties": {}}})
        schema_dict.update({"mdmFlexible": mdmFlexible})
        schema_dict.update({"mdmExportData": export_data})

        for prop, value in base_object.items():
            schema_dict["mdmStagingMapping"]["properties"].update({prop: _dictConstructor(base_object=value)})

        self.schema_dict = schema_dict
        return schema_dict

    def to_json(self, mdmStagingType='stagingName', mdmFlexible=False, crosswalkname=None,
                crosswalkList=None, export_data=False):
        if self.schema_dict is not None:
            json_schema = json.dumps(self.schema_dict)
        else:
            json_schema = json.dumps(self.to_dict(mdmStagingType=mdmStagingType, mdmFlexible=mdmFlexible,
                                                  crosswalkname=crosswalkname,
                                                  crosswalkList=crosswalkList, export_data=export_data))
        return json_schema


def _dictConstructor(base_object):
    schema_dict = {}
    base_object_type = type(base_object)
    schema_type = Type.get_schema_type_for(base_object_type)

    # schema_dict["required"] = True
    schema_dict["type"] = schema_type.json_type
    # schema_dict["fields"]['raw'] = schema_type.json_type

    if schema_type == ObjectType and len(base_object) > 0:
        schema_dict["properties"] = {}

        for prop, value in base_object.items():
            schema_dict["properties"][prop] = _dictConstructor(base_object=value)

    elif schema_type == ArrayType and len(base_object) > 0:
        first_item = base_object[0]
        first_item_type = type(first_item)
        same_type = all((type(item) == first_item_type for item in base_object))

        if same_type:

            first_item_schema_type = Type.get_schema_type_for(first_item_type)
            schema_dict["type"] = first_item_schema_type.json_type

            if first_item_schema_type == ObjectType:
                schema_dict.update(_dictConstructor(base_object=first_item))

        else:
            schema_dict["properties"] = {}
            #schema_dict['items'] = []

            for idx, item in enumerate(base_object):
                schema_dict['properties'].append(_dictConstructor(base_object=item))

    return schema_dict



"""
One of the main purposes of using pipeline managers like Luigi, Dask, Airflow is
to store the result of heavy computation tasks for further reuse. It happens
often that the process of developing the whole pipeline is very dynamic. The
code of a task can be explicitly changed or the task behavior can change due
to external changes, like an update of a python package. It turns out that
one cannot be certain that the stored result of the task is up-to-date with
the code. This may invalidate any test/validation process.
The solution we propose here is to generate a hash of the code of each task.
In the case of using Luigi, we assume that if the target was generated with the
same task parameters and it has the same hash code we can safely use this
target (of course, excluding hash collision hypothesis). We observe that some
pipeline managers also keep track of the data/inputs hash. This module can
cooperate with these managers as well.
Finally, the practical objective of this module is to obtain a hash code of a
given function statically, i.e., without computing the python code itself.
Due to its statical nature, some case cannot be supported, like import and
functions definitions inside if statements.
The method get_bytecode_tree performs all the hard work and return nested lists
of bytecodes that may be useful for inspection.
The method get_function_hash uses get_bytecode_tree to return a hash for a
given function.
"""

from .hash_versioning import get_bytecode_tree, get_function_hash
from pycarol.pipeline.utils.hash_versioning.inspect_bytecode import \
    get_name_of_CALL_FUNCTION



import json


class DataModelFields(object):
    def __init__(self, carol):

        self.carol = carol
        self.offset = 0
        self.page_size = 100
        self.sort_order = 'ASC'
        self.sort_by = None

        self.fields_dict = {}
        self.fields_data = []

        self.all(admin=True, print_status=False, save_file=False)

    def possible_types(self, admin=True):

        if not admin:
            url_filter = "v1/fields/possibleTypes"
        else:
            url_filter = "v1/admin/fields/possibleTypes"

        result = self.carol.call_api(url_filter)
        self._possible_types = result

    def _build_query_params(self):
        if self.sort_by is None:
            self.query_params = {"offset": self.offset, "pageSize": str(self.page_size), "sortOrder": self.sort_order}
        else:
            self.query_params = {"offset": self.offset, "pageSize": str(self.page_size), "sortOrder": self.sort_order,
                                 "sortBy": self.sort_by}

    def all(self, admin=False, offset=0, page_size=100, sort_order='ASC', sort_by=None, print_status=True,
            save_file=False, filename='data/fields.json'):

        self.offset = offset
        self.pageSize = page_size
        self.sortOrder = sort_order
        self.sortBy = sort_by
        self._build_query_params()

        self.fields_dict = {}
        self.fields_data = []
        count = self.offset

        set_param = True
        self.total_hits = float("inf")
        if save_file:
            file = open(filename, 'w', encoding='utf8')

        if admin:
            url_filter = "v1/admin/fields"
        else:
            url_filter = "v1/fields"

        while count < self.total_hits:

            query = self.carol.call_api(url_filter, params=self.query_params)

            count += query['count']
            if set_param:
                self.total_hits = query["totalHits"]
                set_param = False

            query = query['hits']
            self.fields_data.extend(query)
            self.fields_dict.update({i['mdmName']: i for i in query})
            self.query_params['offset'] = count
            if print_status:
                print('{}/{}'.format(count, self.total_hits), end='\r')
            if save_file:
                file.write(json.dumps(query, ensure_ascii=False))
                file.write('\n')
                file.flush()
        if save_file:
            file.close()

    def get_by_id(self, fields_ids=None, admin=False, save_file=False,
                  filename='data/fields_ids.json'):

        assert fields_ids is not None

        self.fields_data = []

        if isinstance(fields_ids, str):
            fields_ids = [fields_ids]
        assert isinstance(fields_ids, list)

        if save_file:
            file = open(filename, 'w', encoding='utf8')

        for fields_id in fields_ids:
            if admin:
                url_filter = f"v1/admin/fields/{fields_id}"
            else:
                url_filter = f"v1/fields/{fields_id}"

            query = self.carol.call_api(url_filter)

            self.fields_data.extend(query)
            self.fields_dict.update({i['mdmName']: i for i in query})

            if save_file:
                file.write(json.dumps(query, ensure_ascii=False))
                file.write('\n')
                file.flush()
        if save_file:
            file.close()

    def create(self, mdm_name, mdm_mpping_data_type, mdm_field_type, mdm_label,
               mdm_description, admin=False):
        '''
        :param mdm_name:
        :param mdm_mpping_data_type: string,  double, long, stc
        :param mdm_field_type: PRIMITIVE or NESTED
        :param mdm_label:
        :param mdm_description:
        :param admin
        :return:
        '''

        if admin:
            url = "v1/admin/fields"
        else:
            url = "v1/fields"

        payload = {"mdmName": mdm_name, "mdmMappingDataType": mdm_mpping_data_type,
                   "mdmFieldType": mdm_field_type,
                   "mdmLabel": {"en-US": mdm_label}, "mdmDescription": {"en-US": mdm_description}}
        assert not mdm_name in self.fields_dict.keys()
        query = self.carol.call_api(method='POST', path=url, data=payload)

        self.fields_dict.update({mdm_name: query})



import dash
import dash_core_components as dcc
from dash.dependencies import Input,Output,State
import dash_html_components as html
from dash_html_components import Button, Div
import dash_cytoscape as cyto
import json
cyto.load_extra_layouts()
external_stylesheets = ['https://codepen.io/chriddyp/pen/bWLwgP.css']

def get_app_from_pipeline(pipe):
    app = dash.Dash(__name__, external_stylesheets=external_stylesheets)
    dag = pipe.get_dag()

    def get_nodes_elements():
        pipe.update_all_complete_status()
        return [
            {
                'data': {
                    'id': k.task_id,
                    'label': k.get_task_family().split('.')[-1],
                    'complete': pipe.get_task_complete(k),
                    'i': i,
                },
                'classes': k.get_task_family().split('.')[-2],
            }
            for i,k in enumerate(dag.keys())
        ]
    def get_edges_elements():
        return [
            {'data': {'source': source.task_id, 'target': target.task_id, }} for
            i, source in enumerate(dag.keys()) for target in dag[source]
        ]


    nodes_elements = get_nodes_elements()
    edges_elements = get_edges_elements()

    group_selectors = [
        {
            'selector': 'node',
            'style': {
                'font-size': '18',
                'content': 'data(label)',
                'shape': 'ellipse',
                'opacity': 1.0,
                'border-style': 'solid',
                'border-width': 5,

            }
        },

        {
            'selector': 'edge',
            'style': {
                'curve-style': 'bezier',
                'source-arrow-shape': 'vee',
                'arrow-scale': 2,
            }
        }
    ]
    import colorcet as cc

    class_selectors = [
        {
            'selector': '.' + family,
            'style': {
                'background-color': cc.glasbey_dark[i],
                'border-color': cc.glasbey_dark[i],
                'selectable': True,
            }
        }
        for i,family in enumerate(
            {t.get_task_family().split('.')[-2] for t in dag.keys()}
        )
    ]
    complete_selector =[
        {
            'selector': "[!complete]",
            'style': {
                'background-color': 'white',
                'shape': 'round-rectangle',
            }
        }
    ]
    default_stylesheet = group_selectors + class_selectors + complete_selector


    def row(*children):
        return html.Div(className='row',children=list(children))

    def column(*children,size=12,**params):
        class_name_dict = {
            6: "six columns",
            12: "twelve columns",
        }
        return html.Div(className=class_name_dict[size],children=list(
            children),**params)


    main_pipeline = cyto.Cytoscape(
        id='main_pipeline',
        layout=dict(
            name='dagre',
            # rankDir='RL',
            # ranker='longest-path',
            padding=5,
            nodeDimensionsIncludeLabels=True,
            animate=True,
            animationDuration=2000,
        ),
        style={
            'width':'100%',
            'height':'600px',
            'background-color': '#fdfd96',
        },
        stylesheet=default_stylesheet,
        elements=nodes_elements + edges_elements,
    )

    tap_node_data = Div(
        id='tap_node_data_div',
        children=[
            html.P(id="tap_node_header", children="Selected Nodes"),
            html.P(id='tap_node_data'),
        ],
        style={'background-color': '#fdfd96'},
    )

    run_button = Button(id='run_button',children='Run')
    remove_button = Button(id='remove_button',children='Remove')
    remove_upstream_button = Button(id='remove_upstream_button',
                                    children='Remove Upstream')
    update_button = Button(id='update_button',children='Update')


    buttons_list = [update_button,run_button,remove_button,
                    remove_upstream_button]

    buttons_output = [
        html.H3(
            id='buttons_output_header',
            children='Buttons Output',
        ),
        html.P(
            id='buttons_output',
        )
        ]

    selected_output = [
        html.H3(
            id='selected_output_header',
            children='Selected Nodes',
        ),
        html.P(
            id='selected_output',
        ),
        ]


    app.layout = column(
        row(main_pipeline),
        row(
            column(
                row(*buttons_list),
                column(*buttons_output, size=6,),
                size=6,
                style={'background-color': '#fdfd96'},
            ),
            column(*selected_output,size=6,style={'background-color': '#fdfd96'},)
        ),
    )

    ### Dynamics ###
    def select_callback(*ts_list):
        max_ts = -1
        print(ts_list)
        for i,ts in enumerate(ts_list):
            if ts and ts>max_ts:
                max_ts = ts
                cb_i = i
        if max_ts == -1:
            return None
        else:
            return cb_i

    def cb_run(nodes):
        if len(nodes) == 0:
            return "No task selected."
        if len(nodes) > 1:
            return "Please, select only one task to run."
        task_id = nodes[0]['id']
        task = pipe.get_task_by_id(task_id)
        # task.run()
        return f"Fake Run {task_id}"

    def cb_goto(nodes):
        pass

    def cb_remove(nodes):
        for n in nodes:
            task_id = n['id']
            task = pipe.get_task_by_id(task_id)
            task.remove()
        return "Removing:\n" + json.dumps(nodes)

    def cb_remove_upstream(nodes):
        tasks = [pipe.get_task_by_id(n['id']) for n in nodes]
        pipe.remove_upstream(tasks)
        return "Removing upstream from:\n" + json.dumps(nodes)

    def cb_dummy(*args,**kwargs):
        return "Updating complete targets."

    @app.callback(
        Output('buttons_output', 'children'),
        [
            Input(b.id, 'n_clicks_timestamp')
            for b in buttons_list
        ],
        [State(main_pipeline.id, 'selectedNodeData')],
    )
    def callback_buttons(*inputs):
        callback_list = [cb_dummy,cb_run,cb_remove,cb_remove_upstream]
        assert len(callback_list) == len(buttons_list)
        ts_list = inputs[0:-1]
        selected_nodes = inputs[-1]
        cb_i = select_callback(*ts_list)
        if cb_i is None:
            return None
        else:
            return callback_list[cb_i](nodes=selected_nodes)

    @app.callback(
        Output('selected_output', 'children'),
        [Input(main_pipeline.id, 'selectedNodeData')],
    )
    def cd_select_node(nodes):
        output = []
        if not nodes:
            return "No task selected."
        for n in nodes:
            task_id = n['id']
            task = pipe.get_task_by_id(task_id)
            task_address = task.get_task_address()
            if task_address is None:
                output.append(task.get_task_family())
            else:
                output.append(
                    dcc.Link(
                        task.get_task_family(),
                        href='/tree/' + task_address,
                    )
                )
            output.append(" ")
        return output

    @app.callback(
        Output(main_pipeline.id, 'elements'),
        [Input(update_button.id, 'n_clicks_timestamp')]
    )
    def cb_update(ts):
        print("cb_update")
        return get_nodes_elements() + get_edges_elements()

    return app



"""
PyCarol Pipeline
================

This module defines a framework for building and managing data pipelines. It was
inspired in Luigi's architecture and has it as a backend for the
implementation of the pipeline execution.

In submodule targets, luigi targets classes are implemented. Some of these
targets, make use of Carol Data Storage and are a key feature for running the
pipeline in the cloud.

In submodule tasks, we found Carol's extension to luigi.Task class. This new
class, combined with some decorators, allows a more efficient description of
the pipeline. It is also in this submodule, that we implement the notebook
task feature, which allows to write task definitions directly in jupyter
notebooks.

In submodule tools, we have tools to interact with the pipeline as a whole.
This inserts an abstraction layer over luigi and provides some features that
are not supported in luigi.

An app pipeline is mainly described using classes from submodules tasks and
targets. Finally, the top pipeline tasks together with the pipeline
parameters definitions are used to instantiate an object of class Pipe,
from submodule tools. This object should be used to make all interactions
with the pipeline, like running, removing targets and related actions.

Finally, submodule viewer implements a dash web app to visualize an object Pipe
and, thus, visualize a given pipeline. This web app has also a jupyter
version to be used inside jupyter lab.

Example
-------

This is a very simplified case when we will create a task that will fetch data from carol in one task and process in
a second task.

.. code:: python

    from dotenv import load_dotenv
    load_dotenv(override=True) #load env variable to create the Carol instance
    from pycarol import Carol, Staging
    from pycarol.pipeline import inherit_list, Task
    import luigi

    @inherit_list(
    )
    class Ingestion(Task):

        connector_name = luigi.Parameter()
        staging_name = luigi.Parameter()

        def easy_run(self, inputs):
            staging_name = self.staging_name
            connector_name = self.connector_name
            stag  = Staging(Carol())
            cds=True
            df = stag.fetch_parquet(staging_name=staging_name, connector_name=connector_name,cds=cds )
            return df


    @inherit_list(
        Ingestion,
    )
    class DataProcess(Task):

        connector_name = luigi.Parameter()
        staging_name = luigi.Parameter()

        def easy_run(self, inputs):
            df = inputs[0] #since there is only one requirement.
            ... #any processing
            return df

    task = [DataProcess(connector_name='new_connector', staging_name='iris')]
    luigi.build(task, local_scheduler=True)

Notice that we can start the pipeline with luigi.build. Pycarol has a method for that too.
If  a single task needs to be build, one can call, from the example above,

.. code:: python

    task = DataProcess(connector_name='new_connector', staging_name='iris')
    task.buildme(local_scheduler=True,)

It will have exactly the same behavior.

"""

from .task import (
    Task,
    WrapperTask,
    inherit_list,
    inherit_dict
)

from .targets import (
    CDSTarget,
    PickleTarget,
    KerasTarget,
    DummyTarget,
    JsonTarget,
    FeatherTarget,
    PytorchTarget,
    ParquetTarget,
    FileTarget,
    LocalTarget
)
from .targets.deprecated_targets import (
    PicklePyCarolTarget,
    PytorchPyCarolTarget,
    KerasPyCarolTarget,
    PickleLocalTarget,
    KerasLocalTarget,
    PytorchLocalTarget,
    JsonLocalTarget,
    FeatherLocalTarget,
    ParquetPyCarolTarget
)

from .tools import (
    Pipe,
)


import unittest
import sys
import json
sys.path.append('../.')
from pycarol.schema_generator import carolSchemaGenerator

class SchemaTestCase(unittest.TestCase):
    
    def test_assert(self):

        tweet = {"created_at": "2019-04-06T04:20:41", "id": 1114382394935652352, "id_str": "1114382394935652352",
                 "text": "@YazCrypto If the price goes down.  That will be bearish.  $LINK",
                 "source": "<a href=\"http://twitter.com/download/android\" rel=\"nofollow\">Twitter for Android</a>",
                 "in_reply_to_screen_name": "YazCrypto",
                 "user": {"id": 930238793512292352, "id_str": "930238793512292352", "name": "Alexander",
                          "screen_name": "AlexSau37473923", "location": "Florida, USA", "url": None,
                          "description": "Started with $BTC mining back in 2013.  \n\n\nDiversified into:\n\n$LINK $XMR $DRGN $MAN $STK\n\nLong On Crypto.",
                          "protected": False, "verified": False, "followers_count": 87, "friends_count": 184,
                          "listed_count": 0, "favourites_count": 757, "statuses_count": 910,
                          "created_at": "2017-11-14T00:59:46", "utc_offset": None, "time_zone": None,
                          "geo_enabled": False, "lang": "en", "default_profile": True, "default_profile_image": False,
                          "notifications": None}, "geo": None, "coordinates": None, "place": None, "contributors": None,
                 "is_quote_status": False, "quote_count": 0, "reply_count": 0, "retweet_count": 0, "favorite_count": 0,
                 "entities": {"hashtags": [], "user_mentions": [
                     {"screen_name": "YazCrypto", "name": "Yaz", "id": 965803417175142400,
                      "id_str": "965803417175142400", "indices": [0, 10]}],
                              "symbols": [{"text": "LINK", "indices": [59, 64]}]}, "filter_level": "low", "lang": "en",
                 "tweet_url": "https://twitter.com/AlexSau37473923/status/1114382394935652352",
                 "coins_mentioned": ["LINK"], "coins_mentioned_count": 1, "sentiment_score": 0.5944444444444444,
                 "sentiment_score_subjectivity": 0.2888888888888889, "reaction_score": 1.295,
                 "crawler": "lightning-work-01"}

        self.assertEqual(1114382394935652352, tweet['id'])
        self.assertEqual('AlexSau37473923', tweet['user']['screen_name'])

        schema = carolSchemaGenerator(tweet)
        schema = schema.to_dict(mdmStagingType='tweet', mdmFlexible=False, export_data=False,
                                crosswalkname='tweet', crosswalkList=['id'])

        print(json.dumps(schema, indent=4))
        self.assertEqual('tweet', schema['mdmStagingType'])
        self.assertTrue('id' in schema['mdmStagingMapping']['properties'])
        self.assertTrue('id' in schema['mdmStagingMapping']['properties']['user']['properties'])


if __name__ == '__main__':
    unittest.main()
    
    
    

    